/*******************************************************************
 *******************************************************************

 Class: FTT-Ethernet                                  
 Type:  FTT Ethernet example application
 Platform: Shark
 Name: eg_rm_s.c
 
 By: Paulo Pedreiras
 Date: 2003-Ago-1
 Description: Two slave nodes, selected by ID {1,2} in the command line
              Configures a set of synchronous messages, produced and consumed vars
              Shows the use of callbacks (both produce and consume)
              Uses Rate Monotonic scheduler (at the master node)
              Slave component

 Log: (Date,Author,description)Comments:

 ********************************************************************
 ********************************************************************/


#include "ftts1_0.h"

  /*****************************/
 /* Application related stuff */
/*****************************/

/* Mutex to serialize user messages */
/* Note: the cprintf() function is not reentrant (since it changes the cursor position) */
/*       so the output becomes messy if tasks using cprintf() are activated when others */
/*       also using cprintf() are executing. */ 
mutex_t umesg_access_mutex;  

int stationid; /* Selects between station 1 or 2 */


    /*****************************************************/
   /*****************************************************/
  /*           User functions                          */
 /*****************************************************/
/*****************************************************/

TASK produce_callback(void)
{
  int i;
  unsigned char mesg_data[100],valor;
  
  valor = 0;
  
  while(1)
    {
      mutex_lock(&umesg_access_mutex);
      
      cprintf("\n\n Locally produced synchronous message txmitted! (First 2 bytes: [%d] [%d])", \
		  mesg_data[0], mesg_data[1]);
      
      mutex_unlock(&umesg_access_mutex);

      /* Init data vector */
      valor++;
      for(i=0;i<10;i++)
	mesg_data[i]=valor+i;
    
      /* Update buffer */
      if(stationid == 1)
	produce(1,mesg_data);

      if(stationid == 2)
	produce(4,mesg_data);

      task_endcycle();
    }
  
}

TASK consume_callback(void)
{
  int errcode;
  BYTE mesg_data[1024];
  long deadline;
  int i;

  while(1)
    {
      mutex_lock(&umesg_access_mutex);
      cprintf("\n\n Locally consumed synchronous message received!");

      if(stationid == 1)
	{
	  if((errcode=consume(4,mesg_data,&deadline)) >= 0)
	    cprintf("( deadline=%ld)\n",deadline);
	  else
	    cprintf("\n Receive error! error code %d",errcode);
	}
      
      if(stationid == 2)
	{
	  if((errcode=consume(1,mesg_data,&deadline)) >= 0)
	    cprintf("( deadline=%ld)\n",deadline);
	  else
	    cprintf("\n Receive error! error code %d",errcode);
	}

      for(i=0;i<10;i++)
	cprintf("%3u ",mesg_data[i]);
      
      mutex_unlock(&umesg_access_mutex);


      task_endcycle();
    }
  
}


    /*****************************************************/
   /*****************************************************/
  /*           Main                                    */
 /*****************************************************/
/*****************************************************/

int main(int argc, char **argv)
{
  
  /* General variables */
  NRDB_mesgtype aux_var;
  signed char vstatus;
  NOP_mutexattr_t umesg_access_mutex_m; /* Mutex to serialize user messages */
  
  
  /* Callback tasks */
  SOFT_TASK_MODEL produce_callback_tm; 
  PID pid_produce_callback;

  SOFT_TASK_MODEL consume_callback_tm;  
  PID pid_consume_callback;

  //set_exchandler_grx(); /* After execption prints PID ... */

  /*   FTT global vars init      */
  EC_LEN=20000; /* EC length (us) */

  /* Check input arguments */
  if(argc != 2)
    {
      cprintf("\n\n Invalid number of arguments");
      cprintf("\n Usage: eg_rm_s ID   (ID={1,2})");
      sys_end();
    }

  stationid=atoi(argv[1]);
  if(stationid != 1 && stationid != 2)
    {
      cprintf("\n\n Invalid station ID!");
      cprintf("\n Usage: eg_rm_s ID   (ID={1,2})");
      sys_end();
    }

  /* Init the ftt system */
  ftt_ninit();

   /************************/
  /* Initialize NOP mutex */
  /************************/
  NOP_mutexattr_default(umesg_access_mutex_m);
  mutex_init(&umesg_access_mutex,&umesg_access_mutex_m);
  

    /**************************************/
   /* Set-up the message set             */
  /**************************************/

  if(stationid == 1)
    {
      
      /* STA1 : Message set */
      
      /*******************************/
      /* Create callback tasks       */
      /*******************************/
      soft_task_default_model(produce_callback_tm);
      soft_task_def_met(produce_callback_tm,2000);
      soft_task_def_period(produce_callback_tm,EC_LEN/2);
      soft_task_def_aperiodic(produce_callback_tm);
      soft_task_def_level(produce_callback_tm,3);   
      if ((pid_produce_callback=task_create("produce_callback", produce_callback, &produce_callback_tm, NULL)) == -1)
	{
	  perror("[eg_rm_s.c(main)] Could not create <produce_callback>:");
	  sys_end();
	}
      cprintf("\n[main] <produce_callback> task created. PID=%u",(unsigned int)pid_produce_callback);
      
      soft_task_default_model(consume_callback_tm);
      soft_task_def_met(consume_callback_tm,2000);
      soft_task_def_period(consume_callback_tm,EC_LEN/2);
      soft_task_def_aperiodic(consume_callback_tm);
      soft_task_def_level(consume_callback_tm,3);   
      if ((pid_consume_callback=task_create("consume_callback", consume_callback, &consume_callback_tm, NULL)) == -1)
	{
	  perror("[eg_rm_s.c(main)] Could not create <consume_callback>:");
	  sys_end();
	}
      cprintf("\n[main] <consume_callback> task created. PID=%u",(unsigned int)pid_consume_callback);
      
      
      SET_MESG_PROP(&aux_var,1,100,1,SDATA_MESG_ID,PRODUCER,pid_produce_callback); /*id,size,deadline,type,status,pidcallback*/
      vstatus=NRDB_addmesg(&aux_var);
      if(vstatus != 0) cprintf("\n [main]: Error returned by NRDB_addmesg (Mesg ID:%x, Err code:%d)\n",aux_var.m_id,vstatus);
      
      SET_MESG_PROP(&aux_var,3,1024,2,SDATA_MESG_ID,PRODUCER,0); /*id,size,deadline,type,status,pidcallback*/
      vstatus=NRDB_addmesg(&aux_var);
      if(vstatus != 0) cprintf("\n [main]: Error returned by NRDB_addmesg (Mesg ID:%x, Err code:%d)\n",aux_var.m_id,vstatus);
      
      SET_MESG_PROP(&aux_var,2,100,1,SDATA_MESG_ID,CONSUMER,0); /*id,size,deadline,type,status,pidcallback*/
      vstatus=NRDB_addmesg(&aux_var);
      if(vstatus != 0) cprintf("\n [main]: Error returned by NRDB_addmesg (Mesg ID:%x, Err code:%d)\n",aux_var.m_id,vstatus);
      
      SET_MESG_PROP(&aux_var,4,1024,5,SDATA_MESG_ID,CONSUMER,pid_consume_callback); /*id,size,deadline,type,status,pidcallback*/
      vstatus=NRDB_addmesg(&aux_var);
      if(vstatus != 0) cprintf("\n [main]: Error returned by NRDB_addmesg (Mesg ID:%x, Err code:%d)\n",aux_var.m_id,vstatus);
      
    }

  if(stationid == 2)
    {
      
      /* STA2 : Message set */
      
      /*******************************/
      /* Create callback tasks       */
      /*******************************/
      soft_task_default_model(produce_callback_tm);
      soft_task_def_met(produce_callback_tm,2000);
      soft_task_def_period(produce_callback_tm,EC_LEN/2);
      soft_task_def_aperiodic(produce_callback_tm);
      soft_task_def_level(produce_callback_tm,3);   
      if ((pid_produce_callback=task_create("produce_callback", produce_callback, &produce_callback_tm, NULL)) == -1)
	{
	  perror("[eg_rm_s.c(main)] Could not create <produce_callback>:");
	  sys_end();
	}
      cprintf("\n[main] <produce_callback> task created. PID=%u",(unsigned int)pid_produce_callback);
      
      soft_task_default_model(consume_callback_tm);
      soft_task_def_met(consume_callback_tm,2000);
      soft_task_def_period(consume_callback_tm,EC_LEN/2);
      soft_task_def_aperiodic(consume_callback_tm);
      soft_task_def_level(consume_callback_tm,3);   
      if ((pid_consume_callback=task_create("consume_callback", consume_callback, &consume_callback_tm, NULL)) == -1)
	{
	  perror("[eg_rm_s.c(main)] Could not create <consume_callback>:");
	  sys_end();
	}
      cprintf("\n[main] <consume_callback> task created. PID=%u",(unsigned int)pid_consume_callback);
      
      
      SET_MESG_PROP(&aux_var,1,100,1,SDATA_MESG_ID,CONSUMER,pid_consume_callback); /*id,size,deadline,type,status,pidcallback*/
      vstatus=NRDB_addmesg(&aux_var);
      if(vstatus != 0) cprintf("\n [main]: Error returned by NRDB_addmesg (Mesg ID:%x, Err code:%d)\n",aux_var.m_id,vstatus);
      
      SET_MESG_PROP(&aux_var,3,1024,2,SDATA_MESG_ID,CONSUMER,0); /*id,size,deadline,type,status,pidcallback*/
      vstatus=NRDB_addmesg(&aux_var);
      if(vstatus != 0) cprintf("\n [main]: Error returned by NRDB_addmesg (Mesg ID:%x, Err code:%d)\n",aux_var.m_id,vstatus);
      
      SET_MESG_PROP(&aux_var,2,100,1,SDATA_MESG_ID,PRODUCER,0); /*id,size,deadline,type,status,pidcallback*/
      vstatus=NRDB_addmesg(&aux_var);
      if(vstatus != 0) cprintf("\n [main]: Error returned by NRDB_addmesg (Mesg ID:%x, Err code:%d)\n",aux_var.m_id,vstatus);
      
      SET_MESG_PROP(&aux_var,4,1024,5,SDATA_MESG_ID,PRODUCER,pid_produce_callback); /*id,size,deadline,type,status,pidcallback*/
      vstatus=NRDB_addmesg(&aux_var);
      if(vstatus != 0) cprintf("\n [main]: Error returned by NRDB_addmesg (Mesg ID:%x, Err code:%d)\n",aux_var.m_id,vstatus);
      
    }


  /* Print station and message set information */
  cprintf("\n-----------------------------");
  if(stationid == 1)
    cprintf("\n    Station 1 ");
  
  if(stationid == 2)
    cprintf("\n    Station 2 ");
  cprintf("\n-----------------------------");

  /* Print additional information (message tables) */
  cprintf("\n Any key to see the message tables");
  keyb_getchar();

  /* Print message tables */
  NRDB_printmesg();
  
  cprintf("\n Any key to continue");
  keyb_getchar();

  
  /* Start the ftt system */
  ftt_nstart();
  
  cprintf("\n System running...");
  
  
  
    /*********************/
   /* *** Test code *** */
  /*********************/
  
  if(stationid == 1)
    {
      unsigned char mesg_data[100];
      int j;
      
      /* Init data vector */
      for(j=0;j<100;j++)
	mesg_data[j]=0x55;
      
      /* Update buffer */
      /* When message is txmitted, the respective callback */
      /* will change the data                              */
      produce(1,mesg_data);
      
      /* Forever */
      while(1)
	{
	  j=j;
	}
    }


  if(stationid == 1)
    {
      unsigned char mesg_data[1024];
      int j;
      
      /* Init data vector */
      for(j=0;j<1024;j++)
	mesg_data[j]=0x55;
      
      /* Update buffer */
      /* When message is txmitted, the respective callback */
      /* will change the data                              */
      produce(4,mesg_data);
      
      /* Forever */
      while(1)
	{
	  j=j;
	}
    }
  
  return 0;
  
}

