/*********************************************************************************************
 ********************************************************************************************/
/*! \file ftti1_0.h
\brief Register Shark modules (scheduling and others) and set-up of network

 Register Shark modules (scheduling and others) and set-up of network\n
 Class: FTT-Ethernet\n
 Type:  FTT Ethernet init file (both Master and Slave)\n
 Platform: Shark\n
 Name: ftti1_0.h\n
 
 By: Paulo Pedreiras\n
 Date: 2003-Jul-31
 
 ***********************************************************************************************
 ***********************************************************************************************/


#include "kernel/kern.h"

/* Scheduling and resource modules */
#include "modules/edf.h"
#include "modules/rm.h"
#include "modules/rrsoft.h"
#include "modules/cbs.h"
#include "modules/rr.h"
#include "modules/dummy.h"

#include "modules/sem.h"
#include "modules/hartport.h"
#include "modules/cabs.h"
#include "drivers/keyb.h"
#include "modules/nop.h"

#include "kernel/mem.h"
#include "ll/i386/cons.h"
#include "drivers/udpip.h"

#include <stdlib.h>
#include <string.h>

#include "drivers/eth.h"


/* System tick in us */
#define TICK 0

/* RR Tick in us */
#define RRTICK 1000


TIME __kernel_register_levels__(void *arg)
{

  struct multiboot_info *mb = (struct multiboot_info *)arg;
  
  /**************************************/
  /* Registration of scheduling modules */
  /**************************************/

  /* ***** Levels used by high-priority tasks ***** */
  /* EDF supports HARD periodic and sporadic tasks  */
  /* Note : this instance of EDF will be used just to register the most important */
  /*        tasks concerning timeliness. (ftt_synmesgprod for Stations and ftt_dispatcher for Master */ 
  EDF_register_level(EDF_DISABLE_ALL);  	
  
  /* CBS supports SOFT periodic and sporadic tasks*/
  /* This level will be used by the ethernet driver */
  CBS_register_level(CBS_DISABLE_ALL,0);
	
  /* ***** General levels for remaining system tasks and user tasks ***** */ 
  /* EDF supports HARD periodic and sporadic tasks*/
  EDF_register_level(EDF_DISABLE_ALL);
  
  /* CBS supports SOFT periodic and sporadic tasks*/
  CBS_register_level(CBS_DISABLE_ALL,2);
	
  /* RR module. Handles the main() task */
  RR_register_level(RRTICK,RR_MAIN_YES, mb);
  
  /* Dummy module. Handles the dummy task */
  /* Required if cant be guaranteed that there is always a ready task */
  dummy_register_level();
  
  /**************************************/
  /* Registration of resource modules   */
  /**************************************/
  
  /* Semaphor module */
  SEM_register_module();

  /* Cyclical Asynchronous Buffer module */
  CABS_register_module();

  /* Register NO protocol (NOP)  resource access protocol (required by mutex) */
  /* I use NOP because I only want to serialize 2 tasks, therefore no priority */
  /* inversions and so on can occur */
  NOP_register_module();
  
  /***********************************************/
  /* End of function: return system tick (in us) */
  /***********************************************/
  return TICK;
  
}

TASK __init__(void *arg)
{
  KEYB_PARMS k;
  struct multiboot_info *mb = (struct multiboot_info *)arg;

  SOFT_TASK_MODEL eth_handler_tm, s; /* Task model for the Ethernet handler and keyboard */
  
  /******************************/
  /* System initialisations     */
  /******************************/
  
  HARTPORT_init();

  soft_task_default_model(s);
  soft_task_def_wcet(s,2000);
  soft_task_def_period(s,25000);     
  soft_task_def_met(s,800);   
  soft_task_def_aperiodic(s);
  soft_task_def_system(s);
  soft_task_def_nokill(s);
  soft_task_def_level(s,3);         /* Register task at level 3 (CBS) */ 

  keyb_default_parm(k);
  keyb_def_task(k, (TASK_MODEL *)&s);
  KEYB_init(&k);
	
  //set_exchandler_grx(); /* After execption prints PID ... */
	
  /*********************************/
  /* Network set-up                */
  /*********************************/
  soft_task_default_model(eth_handler_tm);
  soft_task_def_wcet(eth_handler_tm,1000);
  soft_task_def_period(eth_handler_tm,5000);     /* Minimum ftt frame size is about 100us */
  soft_task_def_met(eth_handler_tm,1000);   
  soft_task_def_aperiodic(eth_handler_tm);
  soft_task_def_system(eth_handler_tm);
  soft_task_def_nokill(eth_handler_tm);
  soft_task_def_level(eth_handler_tm,1);         /* Register task at level 1 (CBS) */ 

  eth_init(0,(TASK_MODEL *)&eth_handler_tm);   /* Init Ethernet interface */
  
  /*********************/
  /* Call main()       */
  /*********************/

  __call_main__(mb);
  
  return (void *)0;

}

