/*********************************************************************************************
 *********************************************************************************************/
/*! \file fttm1_0.h
    \brief Macros, global vars and function prototypes related with the FTT-Ethernet
              Master implementation

 Macros, global vars and function prototypes related with the FTT-Ethernet
              Master implementation
 Class: FTT-Ethernet                                  
 Type:  FTT Ethernet Master master code
 Platform: Shark
 Name: fttm1_0.h
 
 By: Paulo Pedreiras
 Date: 2003-Jul-31 */

 /**********************************************************************************************
 ***********************************************************************************************/

  /********************************/
 /* Includes 			 */
/********************************/

#include <kernel/kern.h>
#include <time.h>
#include <modules/sem.h>
#include <modules/hartport.h>
#include <modules/cabs.h>
#include <drivers/keyb.h>
#include <string.h>
#include <math.h>
#include "modules/nop.h"

#include "kernel/mem.h"
#include "ll/i386/cons.h"

#ifdef ETM_MEASURE
#include "ll/i386/x-dos.h"  /* Required to use the DOS filesystem */
#endif

#include "drivers/udpip.h"

#include "drivers/eth.h"
#include "netbuff.h"

#include "fttps1.h"

#ifndef __FTT_MST_HEADER__
#define __FTT_MST_HEADER__

/**********************************/
/* Macros                         */
/**********************************/

/*! 
\brief Asyncronous message properties
*/
#define SET_AMESG_PROP(mp,mid,msize,mperiod,mdeadline,minit)\
       (mp)->am_id          = (mid),       \
       (mp)->am_size        = (msize),     \
       (mp)->am_period      = (mperiod),   \
       (mp)->am_deadline    = (mdeadline), \
       (mp)->am_init        = (minit)

#ifdef ELASTIC_TM
/*! 
\brief Syncronous message properties

Note: If ELASTIC_TM is not defined, SET_SMESG_PROP becomes\n

SET_SMESG_PROP(mp,mid,msize,mperiod,mdeadline,minit)\n
       (mp)->sm_id          = (mid),       \n
       (mp)->sm_size        = (msize),     \n
       (mp)->sm_period      = (mperiod),   \n
       (mp)->sm_deadline    = (mdeadline), \n
       (mp)->sm_init        = (minit)
*/
  #define SET_SMESG_PROP(mp,mid,msize,mperiod,mperiod_min,mperiod_max,melastic_coef,mdeadline,minit)\
       (mp)->sm_id           = (mid),          \
       (mp)->sm_size         = (msize),        \
       (mp)->sm_period       = (mperiod),      \
       (mp)->sm_period_min   = (mperiod_min),  \
       (mp)->sm_period_max   = (mperiod_max),  \
       (mp)->sm_elastic_coef = (melastic_coef),\
       (mp)->sm_deadline     = (mdeadline),    \
       (mp)->sm_init         = (minit)
#else
  #define SET_SMESG_PROP(mp,mid,msize,mperiod,mdeadline,minit)\
       (mp)->sm_id          = (mid),       \
       (mp)->sm_size        = (msize),     \
       (mp)->sm_period      = (mperiod),   \
       (mp)->sm_deadline    = (mdeadline), \
       (mp)->sm_init        = (minit)
  
#endif       


  /********************************/
 /* Global declarations		     */
/********************************/  

/*!
\brief EC length (us)
*/
unsigned long EC_LEN;  

/*!
\brief SRDB SRT Table, which holds the synchronous message properties
*/
SRDB_SRT_tabtype SRDB_SRT_mesgtable;  

/*!
\brief SRDB ART Table, which holds the synchronous message properties
*/
SRDB_ART_tabtype SRDB_ART_mesgtable;  

/*!
\brief Sequence number of the trigger message
*/
BYTE tm_seqn;                       


/* SRT auxiliary tables used by ETM */

#ifdef ELASTIC_TM

  /*!
  \brief Copy of SRT, only if ELASTIC_TM is defined
  */
  SRDB_SRT_tabtype SRDB_SRT_mesgtable_tmp; 

  /*!
  \brief Copy of SRT for use inside compress, only if ELASTIC_TM is defined
  */
  SRDB_SRT_tabtype SRDB_SRT_mesgtable_ct;  
#endif


/* QUEUES used by EDF scheduler */
#ifdef EDF_SCHED

  /*!
  \brief QUEUES used by EDF scheduler, only if EDF_SCHED is defined
  */
  EDF_READY_QUEUE EDF_rq, EDF_rqtmp; 
#endif


/* CBS data */
#ifdef CBS_SERVER
  CBS_list_type CBS_list; /* Defines a data structure to hold CBSs */

/* Variables used by the ftt_scheduler */
WORD SCHED_CBS_mlist[EC_MAX_VAR]; /* List of ids of messages scheduled for tramsmission in one EC */
int SCHED_CBS_nm=0;               /* Number of messages scheduled for tramsmission in one EC     */
#endif

/*!
\brief Mutex to serialize access to SRT among tasks that use it
*/
mutex_t srt_access_mutex;  

/*!
\brief Allocate mem for 2 ECs. This way, while the scheduler builds one EC, the dispatcher processes the other 
*/
EcType EC_1,  EC_2; 

/*!
\brief Pointer to EC-schedule registers in use
*/
EcType *sched_ec,
  *disp_ec,
  *temp_ec;
 
/*!
\brief Process ID of scheduler and dispatcher tasks
*/
PID pid_ftt_dispatcher,
  pid_ftt_scheduler;    

/*!
\brief Variable shared by ftt_init() and ftt_start()
*/
TIME disp_mit,disp_wcet,sched_mit,sched_wcet;

/*!
\brief Pool of Eth TX buffers
*/
struct netbuff Pool_TxBuff; 


  /********************************************/
 /* Special declarations; debug and measures */
/********************************************/
#ifdef ETM_MEASURE
  /* Structures to record ocurrences of the mesgs */
  #define EVT_TAB_maxevt 3000 /* EC_LEN=20ms -> 50 mesg instances/sec (max) : 60s*50mesg_inst/sec=3000 */
  #define EVT_MAXACT 100      /* Number of actions performed over the SRT */
  long evt_vec[MAX_VAR];
  TIME evt_tab[MAX_VAR][EVT_TAB_maxevt];

  /* Actions performed in the SRT */
  typedef struct
  {
    TIME inst;   /* Instant of event in usec */
    WORD id;     /* Id of message */
    WORD action; /* Action performed 1-delete;2-add new */
  }TSRTACTION;

  TSRTACTION evt_act[EVT_MAXACT];
  int etoact; /* Total of actions performed */
#endif



/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Funcion prototypes >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/  



    /*****************************************************/
   /*****************************************************/
  /*     CBS handling functions                        */
 /*****************************************************/
/*****************************************************/

#ifdef CBS_SERVER

void CBS_init(CBS_list_type *CBS_list);/* Initializes the CBS data structures    */
int CBS_getUcbs(CBS_list_type *CBS_list,double * Ucbs); /* Returns the utilization of the CBS servers currently defined */
int CBS_addserver(CBS_list_type *CBS_list, WORD id, WORD Q, WORD T );/* Add CBS server to the CBS servers list */
int CBS_removeserver(CBS_list_type *CBS_list, WORD id );/* Removes a CBS server from the CBS servers list */
void CBS_printservers(CBS_list_type *CBS_list);/* This function prints the defined CBS servers properties */
int CBS_request(CBS_list_type *CBS_list, WORD id, WORD dlc); /* Handles a request for transmission made to a CBS server  */

#endif /* CBS_SERVER */


    /*****************************************************/
   /*****************************************************/
  /*    Functions used by the Elastic Task Model       */
 /*****************************************************/
/*****************************************************/

#ifdef ELASTIC_TM

signed char ETM_compress(SRDB_SRT_tabtype *srt,double Ud);/* Implements the compress/decompression of a message set */
signed char ETM_changeQOS(SRDB_SRT_mesgtype *aux_var); /* Changes the QOS of existing message and updates Master variables' table */
  
#endif 


    /*****************************************************/
   /*****************************************************/
  /*    SRDB handling functions                        */
 /*****************************************************/
/*****************************************************/

int SRDB_SRT_GetUf(SRDB_SRT_tabtype *srt,double *Uact,double *Ulub);/* Returns the utilization factor of the current  message set  */
int SRDB_SRT_sorttable(); /* Sorts the SRT message table ( Required by RM scheduler ) */
signed char SRDB_SRT_addmesg(SRDB_SRT_mesgtype *aux_var); /*  Appends a message to the master variables' Synchronous Requirements Table */
signed char SRDB_SRT_delmesg(WORD vid); /*  Deletes a synchronous message from the masters SRT */
signed char SRDB_SRT_printmesg(); /*  Prints the Synchronous Requirements Table (For debug!) */

signed char SRDB_ART_addmesg(SRDB_ART_mesgtype *aux_var, unsigned int msg_type, unsigned char timeliness_constr); /*  Appends a message to the master Asynchronous Requirements Table */
signed char SRDB_ART_delmesg(WORD vid); /* Deletes an asynchronous message from the masters ART */
signed char SRDB_ART_printmesg(); /* Prints the Asynchronous Requirements Table (For debug!) */


    /******************************************/
   /******************************************/
  /* EDF ready queue manipulation functions */
 /******************************************/
/******************************************/

#ifdef EDF_SCHED

int EDF_RQ_init(EDF_READY_QUEUE *ready_queue); /*Init EDF ready queue data structures */
int EDF_RQ_insert(EDF_READY_QUEUE *ready_queue, EDF_READY_QUEUE_RECORD new_rec); /* Inserts a new record at the end of the ready queue */
int EDF_RQ_sort_compar(const void *rec1, const void *rec2); /* Compare function used by qsort()*/
int EDF_RQ_sort(EDF_READY_QUEUE *ready_queue); /* Sorts the EDF ready message queue */
int EDF_RQ_update_deadlines(EDF_READY_QUEUE *ready_queue, unsigned int *md_id); /* Updates absolute deadlines of ready messages in the ready queue */
int EDF_RQ_remove_released(EDF_READY_QUEUE *ready_queue, EDF_READY_QUEUE *ready_queue_tmp); /* Removes the released messages from the EDF ready queue */
int EDF_RQ_printqueue(EDF_READY_QUEUE *ready_queue); /* Prints the EDF ready queue contents (for debug)*/

#endif

    /*****************************************************/
   /*****************************************************/
  /*    FTT Tasks implementaion                        */
 /*****************************************************/
/*****************************************************/

TASK ftt_scheduler(void *arg);/* ftt_scheduler() task */
signed char SRDB_printsched(EcType *ec); /* This function prints the EC-schedule contents */
TASK ftt_dispatcher(void *arg); /* Task is invoqued periodically (at the beginning of each EC).*/
void ftt_ethproc(void *eth_pkt); /* Function triggered by the reception of FTT Ethernet Frame */


    /*****************************************************/
   /*****************************************************/
  /*    General functions                              */
 /*****************************************************/
/*****************************************************/

#ifdef ETM_MEASURE
void save_etmtstres(void *arg); /* Function called when the system returns to real mode. Saves measure results */
#endif    


    /*****************************************************/
   /*****************************************************/
  /*    Boot functions                                 */
 /*****************************************************/
/*****************************************************/
int ftt_minit(); /* Function  responsible for initializing the ftt system */
int ftt_mstart(); /* Function responsible for starting the system. Note: ftt_init must be called before this function */

#endif
/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/

