/*******************************************************************
 ******************************************************************/
/*! \file fttps1.h
\brief Global definitions and data structures of FTT, Both Master and Slave component

 Global definitions and data structures of FTT, Both Master and Slave component\n
 Class: FTT-Ethernet\n                                  
 Type:  FTT Ethernet header file\n
 Platform: Shark\n
 Name: fttps1.h\n
 
 By: Paulo Pedreiras\n
 Date: 2001-Nov
 

 Addressing Scheme:\n

	Messages have a 16 bit ID:\n
		- ID 0 is reserved\n
		- The upper 4 bits identify the message type;\n
		  The 5th Msb conveyes the TXNEWDAT flag (for the case of data messages);\n
			therefore valid IDs are in the range 0x0001..0x07FF (1..2047d)\n


			(xxxx)  |  (x)  xxx  |  xxxx  |  xxxx\n
			 \  /      \ /  \-------------------/\n
								  
			  Message   TXNEWDAT message ID\n
			  Type     	bit

*****************************************************************************/

/*! @name General masks 
*/
//@ {
/*! 
\brief Mask for the ID bits of identifier
*/
#define MESGID_MASK	0x07FF

/*! 
\brief Mask for the message type bits
*/
#define MESGTYPE_MASK   0xF000

/*! 
\brief Mask for the 11th bit of identifier (TX_NEW_DAT bit)
*/
#define TXNEWDAT_MASK 	0x0800 	

/*! 
\brief Mask for the TM sequence number in the flags field
*/
#define TMSEQN_MASK     0x00FF  

/*! 
\brief Mask for the Sync. Data message  sequence number in the flags field
*/
#define SDMSEQN_MASK    0x00FF

/*! 
\brief Mask for the Async. Data message sequence number in the flags field
*/
#define ADMSEQN_MASK    0x00FF  

//@}

/*! @name Message type headers
*/
//@ {

/*! 
\brief Master message id header
*/
#define MST_MESG_ID 	0xF000

/*! 
\brief Synchronous data messages id header
*/
#define SDATA_MESG_ID 	0x7000

/*! 
\brief Asynchronous data messages id header - Short data message
*/
#define ADATA_MESG_ID_DATA_SHORT   0x0000 

/*! 
\brief Asynchronous data messages id header - Alarm message
*/
#define ADATA_MESG_ID_ALARM        0x1000 

/*! 
\brief Asynchronous data messages id header - Signal message
*/
#define ADATA_MESG_ID_SIGNAL       0x2000 

/*! 
\brief Asynchronous data messages id header - Long data message
*/
#define ADATA_MESG_ID_DATA_LONG    0x3000 
//@ }

/*! @name Message subtypes
*/
//@ {

/*! 
\brief SIGNAL Assynchronous message carries a CBS Request
*/
#define SIGNAL_CBS_REQ_TYPE        0x01  


/* Message timeliness requirements */

/*! 
\brief No   timeliness constraints
*/
#define TMLN_UNCONST  0    

/*! 
\brief Hard timeliness constraints
*/
#define TMLN_HARD     1    

/*! 
\brief Soft (best effort) timeliness constraints
*/
#define TMLN_SOFT     2    
//@ }

/*! @name Maximum sizes
*/
//@ {

/*! 
\brief Maximum valid ID of messages
*/
#define MAX_ID 		 0x07FF  

/*! 
\brief Maximum number of messages
*/
#define MAX_VAR		 64	 

/*! 
\brief Maximum # of messages in 1 EC (Synchronous and Asynchronous)
*/
#define EC_MAX_VAR  	 100   	 

/*! 
\brief Maximum number of messages that a Station can handle
*/
#define STA_MAX_VAR	 50      

/*!
\brief Maximum load allowed due to the scheduler (in %)
*/
#define SCHED_LOAD_BOUND 80   	 

/*! 
\brief Size (in # of messages) of each buffer associated with asynch mesgs transmitted or recieved locally by the node 
*/
#define STA_AM_BUFF_SIZE  5

/*! 
\brief Worst case len  of TM: Header(ID+Flags+nmesgs sinc and asinc + ID/Txi of sync. mesgs + ID of asynch mesgs + Prop delay
*/                   
#define TM_LEN_bits (long)(sizeof(FTT_TM_HEADER)+3*EC_MAX_VAR*sizeof(BYTE)+2*EC_MAX_VAR*sizeof(BYTE))*8+ETH_OVERHEAD_bits+FTT_IFS_bits

/*! 
\brief EC length in bits
*/
#define EC_LEN_bits    (long)(EC_LEN/BIT_TIME_US) 

/*! 
\brief Percentage of the EC reserved to the Synch Phase
*/
#define SWIN_SHARE	0.5

/*! 
\brief Percentage of the EC reserved to the Asynch Phase
*/
#define AWIN_SHARE      (float)(1-SWIN_SHARE) /* Percentage of the EC reserved to the Asynch Phase */
                      /* Note : These values should no be used in the code! */
		      /* AWIN_SHARE + SWIN_SHARE cant be 100%, since some space is */
                      /*        used by the trigger message (and other control messages?)*/
                      /*  Use in the code EC_MAX_bits_SP/AP that have already included the overheads */

/*! 
\brief EC Maximum Synchronous Phase duration measured in bits
*/
#define EC_MAX_bits_SP	(long)(EC_LEN_bits*SWIN_SHARE)

/*! 
\brief EC Maximum asynchronous Phase duration measured in bits
*/
#define EC_MAX_bits_AP  (long)(EC_LEN_bits - (EC_MAX_bits_SP + TM_LEN_bits))

/*! 
\brief EC Maximum Synchronous Phase duration (in microsecs)
*/
#define EC_MAX_us_SP	(long)(EC_MAX_bits_SP*BIT_TIME_US) 

/*! 
\brief EC Maximum Asynchronous Phase duration (in microsecs)
*/
#define EC_MAX_us_AP    (long)(EC_MAX_bits_AP*BIT_TIME_US) 

/*! 
\brief Bit time in us (for 10Mb/s Ethernet)
*/
#define BIT_TIME_US   0.1		        

/*! 
\brief Size of the scanning window (in us) after Trig mesg.\n
This time is intended to allow that all stations decode the TM and set the Tx of the first data message
*/
#define SCAN_WINDOW 500  
//@ }

/*! @name Ethernet related constants
*/
//@ {

/*! 
\brief Ethernet Frame type ID
*/
#define ETH_FTT_TYPE 	0x8FF0	

/*! 
\brief Overhead of an Ethernet Frame:\n
 Dest. Addr (6) + Src. Addr (6) + Type (2)\n
 Note : from a logical point of view, Prelude and FCS not included
*/
#define ETH_OVERHEAD	2*6+2	

/*! 
\brief Maximum number of Synchronous Messages allowed in each EC
*/
#define FTT_MAX_SYN_MESGS_EC	64  

/*! 
\brief Worst-case gap that must be forced between short FTT Data frames, due to non-simultaneous\n 
reception of the TM mesg in all nodes and software jitter (on reception of TM / txmit of data frame
*/
#define FTT_IFS_bits	        1024 

/*! 
\brief Minimum allowed size for a data frame in FTT Ethernet\n
(to support bus and software jitter still without collisions)
*/
#define FTT_MINLEN_bits (ETH_OVERHEAD_bits + ETH_DATALEN_MIN*8 + FTT_IFS_bits) 

/*! 
\brief Maximum lenth in bytes of an Eth. frame (complete: headers,FCS...)
*/
#define ETH_MAXLEN_BYTES 1526	     

/*! 
\brief Overhead of Eth. frame:\n
Preamble(7B)+SOF(1B)+Source/dest addr(2*6B)+ Len(2B) + FCS(4B)+ Int. frame gap (9.6us=96bit)=26Bytes + 96 bits
*/
#define ETH_OVERHEAD_bits (26*8+96)  

/*! 
\brief Minimum length of the data field for an Eth. frame.\n
If a frame is shorter, padd is automatically done
*/
#define ETH_DATALEN_MIN	46   

/*! 
\brief Maximum databytes of an Ethernet frame
*/
#define ETH_DATALEN_MAX 1500 

/*! 
\brief Value of the Slot time used as a count unit for message tx time indication in the trig. mesg
*/
#define SLOTV_US    5	

/*! 
\brief Temporal resolution (in usec) of message deadlines
*/
#define DDL_RES_US    1000  

/*! 
\brief Offset for wake-up Eth Tx mesg trig before the correct time in order to turn-around\n
 some odd delays between the sem_post and task execution
*/
#define TXTRIG_OFFSET 200  

/*! 
\brief Overheads in Tx and Rx of message, for time validity computation\n
 Note: Transmission time is !NOT! included here
*/
#define T2DEADLINE_OVERHEADS_US 20 
//@ }

/*! @name Status of the EC-Schedule registers
*/
//@ {

/*! 
\brief EC-schedule built and ready to be used)
*/
#define UPDATED 	0   

/*! 
\brief EC-schedule empty (during boot)
*/
#define OUT_OF_DATE 	1   

/*! 
\brief EC-schedule being used by the scheduler
*/
#define BUSY_SCHED 	2   

/*! 
\brief EC-schedule being used by the dispatcher
*/
#define BUSY_DISP 	3   

/*! 
\brief Hard deadline missed while scheduling
*/
#define MISSED_DEADLINE 4   

/*! 
\brief EC-schedule has been dispatched
*/
#define DISPATCHED      5   
//@ }

/*! @name Scheduler Errors
*/
//@ {

/* Scheduler Errors (err_code field) */
/* Note that a missed deadline in hard messages causes the scheduler to stop */
/* Hard messages are all the synchronous and the hard asynchrobnous */

/*! 
\brief No error
*/
#define SCH_ERRCODE_OK       0 

/*! 
\brief Deadline miss in synchronous message
*/
#define SCH_ERRCODE_SM_MISS_DEADLINE       1 

/*! 
\brief Deadline miss in hard asynchronous message
*/
#define SCH_ERRCODE_AM_HARD_MISS_DEADLINE  2 

/*! 
\brief Deadline miss in soft asynchronous message
*/
#define SCH_ERRCODE_AM_SOFT_MISS_DEADLINE  3 

/*! 
\brief Deadline miss in unconstrained asynchronous message
*/
#define SCH_ERRCODE_AM_UNCON_MISS_DEADLINE 4 
//@ }

/*! @name Message status
*/
//@ {

/*! 
\brief Current station produces this message
*/
#define PRODUCER    	0x01 	

/*! 
\brief Current station consumes this message
*/
#define CONSUMER    	0x02 	

/*! 
\brief =1, value buffer has been refreshed (producer)
*/
#define TX_NEW_DAT  	0x04 	

/*! 
\brief =1, value buffer has been refreshed (consumer)
*/
#define RX_NEW_DAT  	0x08 	
//@ }

#ifdef CBS_SERVER 
/* CBS Servers constants */

/*! 
\brief Maximum number of CBS servers supported
*/
#define CBS_MAXSERV          20 

/*! 
\brief Maximum number of pendinq requests in one server
*/
#define CBS_MAXPENDREQ      100 

/*! 
\brief Data size of a SIGNAL_CBS_REQ asynch. message
*/
#define SIGNAL_CBS_REQ_SIZE   6 
#endif

/*! @name General constants
*/
//@ {

#define TRUE            1
#define FALSE           0

/*! 
\brief Used by the Elastic Task Model - Set feasible
*/
#define ETM_FEASIBLE    1       

/*! 
\brief Set infeasible
*/
#define ETM_INFEASIBLE  0 
//@ }

/*******************************************************************************
 Note that TX_NEW_DAT bit is set (=1) when the value buffer is written
 by the application and is reset (=0) when the message is produced by
 the communication interface (producer station).
    The RX_NEW_DAT bit is set (=1) when the message is consumed by the
 communication interface and is reset (=0) when the respective
 value buffer is read by the application (consumer station).
    If the producing application fails to supply new data at the
 appropriate rate (message period) then the TX_NEW_DAT bit will be reset
 meaning that the same value is being transmitted again.
    If the consuming application reads from the interface faster than the
 message period, then the RX_NEW_DAT bit will remain reset for, at least,
 two consecutive reads meaning that the value has not been updated.
********************************************************************************/




  /***************************/
 /* EDF QUEUE  declarations */
/***************************/

/*! @name EDF QUEUE  declarations
*/
//@ {

/*! 
\brief Maximum number of messages that can be ready simultaneously
*/
#define MAX_READY_MESGS 1000 

/*! 
\brief Record free
*/
#define EDF_RQ_FREE 1

/*! 
\brief Record in use
*/
#define EDF_RQ_BUSY 0 
//@ }


/* EDF ready queue type*/

/*! \struct EDF_READY_QUEUE_RECORD  
\brief EDF ready queue type
*/
typedef struct 
{
  unsigned int  m_id;	     ///< Message id 
  unsigned char m_ssize;     ///< Message size in SLOTS 
  unsigned int  m_deadline;  ///< Message absolute deadline 
  
  char r_free; ///< Free ( != 0) or in use (0) record 
  
} EDF_READY_QUEUE_RECORD;

/*! \struct EDF_READY_QUEUE  
\brief EDF ready queue 
*/
typedef struct
{
  int num_recs; ///< Number of records stored in queue 
  EDF_READY_QUEUE_RECORD rec[MAX_READY_MESGS]; ///< Max ready messages
} EDF_READY_QUEUE;

  /******************************/
 /******************************/
/******************************/




  /*******************************/
 /* Master Data Structures      */
/*******************************/
/*******************************************************/
/* Synchronous Message properties (SRDB_SRT table line)*/
/*******************************************************/

/*! \struct SRDB_SRT_mesgtype  
\brief Master synchronous message type
*/
typedef struct
{
  /* General properties */
  unsigned int  sm_id;	      ///< Message id 
  unsigned int  sm_size;      ///< n. of data bytes 
  unsigned long sm_bsize;     ///< Message size in bits (including all overheads )
  unsigned char sm_ssize;     ///< Message size in SLOTS 
  unsigned int  sm_period;    ///< Message period in n. of ECs 
  unsigned int  sm_deadline;  ///< Message deadline, normally=period 
  signed int    sm_init;      ///< EC of first ocurrence 
  unsigned char sm_status;    ///< Message status: PRODUCER, CONSUMER, TX_NEW_DAT, RX_NEW_DAT 
  unsigned char sm_deleted;   ///< Flag to signal if a message was deleted 

  /* Used by Elastic Task model */
  unsigned int  sm_period_max;    ///< Maximum period of the message 
  unsigned int  sm_period_min;    ///< Minimum period of the message 
  unsigned int  sm_period_nom;    ///< Nominal period of the message 
  float         sm_elastic_coef;  ///< Elastic coefficient (>=0). Greater value means a more "compressible" task 
  unsigned int  sm_period_shadow; ///< Used to store the new period updated after next release 
                                  ///<   (for the case of period decrease). For period increase the change can 
                                  ///<    be made directly on the SRT 
} SRDB_SRT_mesgtype;

/* Master synchronous variables table type */
/*! \struct SRDB_SRT_tabtype  
\brief Master synchronous variables table type
*/
typedef struct
{
  unsigned char sm_nvars;
  SRDB_SRT_mesgtype sm_message[MAX_VAR];
} SRDB_SRT_tabtype;

/********************************************************/
/* Asynchronous Message properties (SRDB_ART table line)*/
/********************************************************/

/*! \struct SRDB_ART_mesgtype 
\brief Master Asynchronous message type
*/
typedef struct
{
  /* General properties */
  unsigned int  am_id;	       ///< Message id 
  unsigned int  am_type;       ///< Message type/subtype 
  unsigned char am_timeliness; ///< Hard/Soft/No constraints 
  unsigned int  am_size;       ///< Number of data bytes 
  unsigned long am_bsize;      ///< Message size in bits (including all overheads )
  unsigned char am_ssize;      ///< Message size in SLOTS 
  unsigned char am_period;     ///< Message period 
  unsigned int  am_deadline;   ///< Message deadline, normally=period 
  signed   int  am_init;       ///< EC of first ocurrence 
  
} SRDB_ART_mesgtype;

/* Master asynchronous variables table type */
/*! \struct SRDB_ART_tabtype 
\brief Master asynchronous variables table type
*/
typedef struct
{
  unsigned char am_nvars;
  SRDB_ART_mesgtype am_message[MAX_VAR];
} SRDB_ART_tabtype;


  /********************************/
 /* Station Data Structures      */
/********************************/
/*******************************************************/
/* Node's Message properties (NRDB_SRT table line)     */
/*******************************************************/

/*! \struct NRDB_mesgtype 
\brief Node's Message properties
*/
typedef struct
{
  unsigned int m_id;          ///< Message id       
  unsigned int m_size;        ///< N. of data bytes 
  unsigned int m_deadline;    ///< Message deadline (in SLOTs) 
  WORD         m_type;        ///< Message type (Synchronous or asynch.(several types e.g. data,signal ...) 
  WORD         m_flags;       ///< Message flags (seq. number + ??                                          
  PID          m_pid_callback;///< Pid of task to activate in Tx (producer) or RX (consumer)                
  short int    m_t2deadline;  ///< Time until deadline (in SLOTs)                                           
  struct timespec m_tupdate;  ///< Time instant of update                                                    
  WORD         m_cbs_smid;    ///< ID of signal message used to send a request to the CBS server            
  WORD         m_nprod;       ///< Number of times that the message has been updated  (for producer nodes)  
  WORD         m_ncons;       ///< Number of times that the message has been read     (for consumer nodes)  
  WORD         m_ntx;         ///< Number of times that the message has been txmitted (for producer nodes)  
  WORD         m_nrx;         ///< Number of times that the message has been received (for consumer nodes)  
  
  /*** Note : communication is done using CABs (synchronous messages) */
  /*          or FIFOs (asynchronous messages)                        */
  /*          To each variable instance will correspond	              */
  /*	      one CAB/FIFO    from a set of STA_MAX_VAR CABs.	      */
  
  unsigned char m_status;   ///< PRODUCER, CONSUMER, TX_NEW_DAT, RX_NEW_DAT 
} NRDB_mesgtype;


/*******************************************************/
/* Node's Message properties table                     */
/*******************************************************/
/*! \struct NRDB_tabtype 
\brief Node's Message properties table
*/
typedef struct
{
  unsigned char m_nmesgs;
  NRDB_mesgtype mesg[STA_MAX_VAR];
} NRDB_tabtype;

/******************************************/
/* Messages to be produced within the EC  */
/* (used internaly in the Slave nodes for */
/* inter-task communication).             */
/******************************************/

/*! \struct STA_M2P_mesgtype 
\brief Produced message type
*/
typedef struct
{
  WORD msg_type;  ///< Synch or Asynch 
  WORD msg_id;    ///< Id of the message 
  BYTE msg_len;   ///< Length of the message (in time slots) 
} STA_M2P_mesgtype;

/*! \struct STA_M2P_tabtype 
\brief Produced message table type
*/
typedef struct
{
  WORD nsm;  ///< Number of Synch mesgs 
  WORD nam;  ///< Number of Asynch mesgs 
  STA_M2P_mesgtype mesg2p[5*EC_MAX_VAR];  ///< Message list  
} STA_M2P_tabtype;



  /***********************/
 /* ECs' type           */
/***********************/
/*! \struct EcType 
\brief ECs' type
*/
typedef struct
{
  unsigned char status;	  ///< Status of the EC-Schedule : UPDATED, OUT_OF_DATE,... 
  unsigned char err_code; ///< Details about (if relevant) the cause of the error 

  unsigned long sm_load;   ///< Current EC load usecs (overheads included) - Synch. mesgs  
  unsigned long am_load;   ///< Current EC load usecs (overheads included) - Asynch. mesgs 
  WORD sm_nmesgs;            ///< Number of synchronous messages in current EC		 
  WORD am_nmesgs;            ///< Number of asynchronous messages in current EC		 
  
  /* Synchronous message data structures */
  WORD EC_SIDlist[EC_MAX_VAR]; ///< List of ID's of sync messages 
                              ///< that should be produced in the current EC 
  
  BYTE EC_SSizelist[EC_MAX_VAR]; ///< Size of the messages produced in the EC (in SLOTS)

  /* Asynchronous messages data structures */
  WORD EC_AIDlist[EC_MAX_VAR]; ///< List of ID's of async messages 
                              ///< that should be produced in the current EC 
} EcType;

  /*******************************************/
 /*   Ethernet Message Definitions          */
/*******************************************/
/*************************************/
/* Synchronous data message contents */
/*********************************************************************/
/* SDM_ID + SDM_Flags + SDM_time2deadline + DataByte[0..n-1]         */
/* (WORD)    (WORD)       (WORD)             n* (byte)               */
/*                                                                   */
/*  SDM_Flags: numero sequencia (SDMSEQN_MASK) + ?                   */
/*********************************************************************/

/*! \struct FTT_SDM_HEADER 
\brief Header data carried in the Synchronous Data Message
*/
typedef struct
{
  WORD SDM_id;	       ///< Frame ID               
  WORD SDM_flags;      ///< Sequence number + ...  
  WORD SDM_t2deadline; ///< Time until deadline     
} FTT_SDM_HEADER;


/**************************************/
/* Asynchronous data message contents */
/*********************************************************************/
/* ADM_ID + ADM_Flags + ADM_time2deadline + DataByte[0..n-1]         */
/* (WORD)    (WORD)       (WORD)             n* (byte)               */
/*                                                                   */
/*  ADM_Flags: numero sequencia (ADMSEQN_MASK) + ?                   */
/*********************************************************************/

/*! \struct FTT_ADM_HEADER 
\brief Header data carried in the Asynchronous Data Message
*/
typedef struct
{
  WORD ADM_id;	       ///< Frame ID               
  WORD ADM_flags;      ///< Sequence number + ...  
  WORD ADM_t2deadline; ///< Time until deadline      
} FTT_ADM_HEADER;


/****************************/
/* Trigger message contents */
/*********************************************************************/
/* TM_ID + TM_Flags + TM_#smesgs + TM_#amesgs                        */
/* (WORD)    (WORD)       (WORD)     (WORD)                          */
/*       + [SMesgID_1 + SMesgLen_1] + ... + [SMesgID_n + SMesgLen_n] */
/*           (WORD)       (BYTE)              (WORD)       (BYTE)    */
/*       + [AMesgID_1] + ... + [AMesgID_m]                           */
/*           (WORD)              (WORD)                              */
/*                                                                   */
/*  TM_Flags: numero sequencia (TMSEQN_MASK) + ?                     */
/*********************************************************************/

/*! \struct FTT_TM_HEADER 
\brief Header data carried in the Trigger Message
*/
typedef struct
{
  WORD TM_id;	 ///< Frame ID              
  WORD TM_flags; ///< Sequence number + ... 
  WORD TM_nsm;   ///< Number of synch. mesgs 
  WORD TM_nam;   ///< Number of asynch mesgs 
  
} FTT_TM_HEADER;

/*! \struct FTT_TM_PKT 
\brief Trigger message contents
*/
typedef struct
{
  FTT_TM_HEADER TM_Header;
  BYTE TM_Data[5*EC_MAX_VAR];	///< Messages that should be produced in each EC.       
                                ///< 3 bytes/each synch mesg + 2 bytes/each asynch mesg 
} FTT_TM_PKT;

/***************************************/
/* CBS Sporadic Server Data Structures */
/*********************************************************************/
/*********************************************************************/

#ifdef CBS_SERVER
/* CBS Server properties */
typedef struct
{
  WORD id;	 /* Message ID                                */
  WORD Q;        /* Server maximum budget                     */
  WORD T;        /* Server period                             */
  int  d;        /* Server current deadline                   */
  WORD c;        /* Server current budget                     */
  WORD n;        /* Server current number of pending requests */
  WORD dlc[CBS_MAXPENDREQ];   /* DLC of each pending request  */ 
 
} CBS_type;

typedef struct
{
  int ns;                     /* Number of active servers */
  CBS_type cbs[CBS_MAXSERV];  /* Set of servers           */
} CBS_list_type;

#endif












