/*******************************************************************
 *******************************************************************/
/*! \file ftts1_0.c
    \brief Function implementation related with the FTT-Ethernet\n
              Slave implementation

    Function implementation related with the FTT-Ethernet\n
              Slave implementation

 Class: FTT-Ethernet\n                                  
 Type:  FTT Ethernet Slave code\n
 Platform: Shark\n
 Name: ftts1_0.c\n
 
 By: Paulo Pedreiras\n
 Date: 2003-Aug-4 
 
 ********************************************************************
 ********************************************************************/

/********************************/
/* Includes 			*/
/********************************/

#include "ftts1_0.h"



    /*****************************************************/
   /*****************************************************/
  /*    NRDB handling functions                        */
 /*****************************************************/
/*****************************************************/


signed char NRDB_addmesg(NRDB_mesgtype *aux_var)
/****************************************************************************/
/*! 
   \brief Inserts a variable in the station variables' table

   Inserts a variable in the station variables' table\n
   The table has a fixed size of STA_MAX_VAR messages.

   \param *aux_var: pointer to struct with properties of variable to add\n
          id\n
          size (data size in bytes)\n
          deadline (in ECs)\n
          status ( producer / consumer)\n
  \param NRDB_tabtype: global variables' table
  \param NRDB_mesgtable: global variables' table
  \return error code\n 
 					 0 : OK\n
                    -1 : Table full (NRDB_mesgtable.m_nmesgs = STA_MAX_VAR)\n
                    -2 : Invalid message ID (null ID)\n
                    -3 : Invalid message id (aux_var->m_id >= MAX_ID)\n
 			        -4 : Invalid message ID (Duplicated ID)\n
      				-5 : No slot available found but it should be one (strange error)\n
                    -6 : Invalid type\n
                    -7 : Invalid message status\n
                    -8 : Invalid deadline\n
                    -9 : invalid SIGNAL ID for sending CBS requests (only if CBS_SERVER defined)
 
**************************************************************************************/
{
  int i;
  unsigned char cabid[16];   /* String with the ID of the CAB */
  unsigned char portid[16];  /* String with the ID of the com. port */
  int cptype;                /* Com. port type (READ or WRITE) */

  /* Long asynch. mesgs */
  unsigned char dl_flag=0;
  WORD mesg_len;
  
  /*****************************/
  /* Check input data validity */
  /*****************************/
  /* Check if table is full; if yes terminate with error */
  if( NRDB_mesgtable.m_nmesgs == STA_MAX_VAR )
    return( -1 );
  
  /* Check if input id==0;if yes terminate with error */
  /* Note that id== 0 identifies an empty slot */
  if( aux_var->m_id == 0 )
    return( -2 );
  
  /* Check if message id is in range of validity  */
  if( aux_var->m_id >= MAX_ID )
    return( -3 );

  /* Check if valid deadline */
  if(aux_var->m_deadline == 0)
    return(-8);

  /* Check for message type validity */
  switch(aux_var->m_type)
    {
    case ADATA_MESG_ID_DATA_SHORT:
    case ADATA_MESG_ID_ALARM:
    case ADATA_MESG_ID_SIGNAL:

    case SDATA_MESG_ID:
      break;

    case ADATA_MESG_ID_DATA_LONG:
      /* Check id of associated SIGNAL message */ 
      if( (aux_var->m_cbs_smid == 0 || aux_var->m_cbs_smid > MAX_ID) && \
	  aux_var->m_status == PRODUCER)
	return(-9);
      
      aux_var->m_cbs_smid |= ADATA_MESG_ID_SIGNAL;
      dl_flag=1;

      break;

    default:
      return(-6);
    }

  /* Check for valid status */
  if(aux_var->m_status != PRODUCER && aux_var->m_status != CONSUMER)
    return(-7);
    
  /* Add header for data message */
  aux_var->m_id |= aux_var->m_type;
  
  /* Check if ID/Type already on the table */
  for(i=0;i<STA_MAX_VAR;i++)
    if(NRDB_mesgtable.mesg[i].m_id==aux_var->m_id)
      return(-4);

  /* Fill remaining fields */
  aux_var->m_t2deadline=0;
  aux_var->m_nprod=0;
  aux_var->m_ncons=0;
  aux_var->m_ntx=0;
  aux_var->m_nrx=0;
  
  /********************/ 
  /* Store input data */
  /********************/
  for(i=0;i<STA_MAX_VAR;i++)	
    { /* Scan var table for 1st empty slot (slot is signalled busy by id!=0)*/
      if(NRDB_mesgtable.mesg[i].m_id == 0)
	{
	  /* Place message properties in table */
	  NRDB_mesgtable.mesg[i] = *aux_var;
	  

	  /* Create CAB or communication port depending on message type */
	  /* For synchronous messages, it is used a CAB (only most recent value is relevant) */
	  /* For asynchronous messages, it is used a com port (all messages are queued) */
	  if(aux_var->m_type == SDATA_MESG_ID)
	    {
	      /* Set-up a CAB for message data manipulation */
	      /* Note that there must be room for n+1 megs if n tasks use the */
	      /* CAB structure */
	      sprintf(cabid,"%x",aux_var->m_id);
	      cabset[i]=cab_create(cabid,aux_var->m_size*sizeof(char),3);
	      
	      //cprintf("\n[NRDB_addmesg] CAB created (Pos= %d, cabid=%s, size=%d)",i,cabid,aux_var->m_size*sizeof(char));
	    }
	  else
	    {
	      if(aux_var->m_status == PRODUCER)
		cptype=READ; /* Station is a producer -> network layers reads the port */
	      else
		cptype=WRITE; /* Station is a consumer -> network layer writes to the port */
	      
	      /* Long asynch mesgs have a dlc field (in the Ethernet frame,first WORD after header) */
	      if(dl_flag)
		mesg_len=(aux_var->m_size)*sizeof(char)+sizeof(WORD); 
	      else
		mesg_len=aux_var->m_size*sizeof(char);

	      /* Each async message holds internally the instant of production and time 2 deadline */
	      mesg_len += sizeof(struct timespec)+sizeof(short int); /* prod. time + time 2 deadline */

	      sprintf(portid,"%x",aux_var->m_id);
	      portset[i]=port_create(portid, mesg_len, STA_AM_BUFF_SIZE, STREAM, cptype);

	      //cprintf("\n[NRDB_addmesg] port created (Pos=%d, portid=%s, size=%d)",i,portid,mesg_len);
	      if(portset[i] < 0)
		cprintf(" !!Error!! creating port!\n");
	      
	    }
	  
	  /* Update number of variables in table */
	  NRDB_mesgtable.m_nmesgs++;

	  /* Exit the function */
	  return(0);
	  
	}
    }
  
  perror("\n [NRDB_addmesg]: Strange error");
  return(-5);
  
}



signed char NRDB_delmesg(WORD MesgID, WORD mesg_type)
/****************************************************************************/
/*!    

\brief Removes a message from the station variables' table

	Removes a message from the station variables' table\n
    The table has a fixed size of STA_MAX_VAR messages.

 \param MesgID: ID of message to remove
 \param mesg_type: type of the message to remove (SDATA_MESG_ID;ADATA_MESG_ID_DATA_SHORT ... )
 \param NRDB_tabtype: global variables' table
 \param NRDB_mesgtable: global variables' table
      
 \returns error code\n
					 0 : OK\n
                    -1 : Message ID not found\n
                    -2 : Invalid message type

**************************************************************************************/
{
  int i;
  
  /* Check for message type validity */
  switch(mesg_type)
    {
    case ADATA_MESG_ID_DATA_SHORT:
    case ADATA_MESG_ID_ALARM:
    case ADATA_MESG_ID_SIGNAL:
    case ADATA_MESG_ID_DATA_LONG:
    case SDATA_MESG_ID:
      break;

    default:
      return(-2);
    }

  /* Build complete id (ID + TYPE) */
  MesgID |= mesg_type;

  /* Check if ID/Type exists on the table */
  for(i=0;i<STA_MAX_VAR;i++)
    if(NRDB_mesgtable.mesg[i].m_id == MesgID)
      {
	/* Remove comm resources allocated to message being deleted */
	if(NRDB_mesgtable.mesg[i].m_type == SDATA_MESG_ID)
	  { /* Remove the CAB associated with the synchronous message */
	    cab_delete(cabset[i]);
	  }
	else
	  { /* Remove the Comm Port associated with the asynchronous message */
	    port_delete(portset[i]);
	  }
	
	/* Remove message from table */
	NRDB_mesgtable.mesg[i].m_id=0;
	NRDB_mesgtable.m_nmesgs--;
	
	/* Return sucess */
	return(0);
      }
  
  /* If message not found, return with error */
  return(-1);
}


signed char NRDB_changemesg(WORD MesgID, WORD mesg_type, WORD size, WORD deadline)
/****************************************************************************/
/*!   
\brief Changes a message from the station variables' table

	Changes a message from the station variables' table\n
    The table has a fixed size of STA_MAX_VAR messages.

\param MesgID: ID of message to remove
\param mesg_type: type of the message to remove (SDATA_MESG_ID;ADATA_MESG_ID_DATA_SHORT ... )
\param size: size of the message
\param deadline: deadline of the message
\param NRDB_tabtype: global variables' table
\param NRDB_mesgtable: global variables' table

\returns error code\n
					  0 : OK\n
                     -1 : Message ID not found\n
                     -2 : Invalid message type\n
                     -3 : Invalid parameters (invalid size and deadline)

**************************************************************************************/
{
  int i;
  unsigned char cabid[16];   /* String with the ID of the CAB */
  unsigned char portid[16];  /* String with the ID of the com. port */
  int cptype;                /* Com. port type (READ or WRITE) */
  
  /* Check for message type validity */
  switch(mesg_type)
    {
    case ADATA_MESG_ID_DATA_SHORT:
    case ADATA_MESG_ID_ALARM:
    case ADATA_MESG_ID_SIGNAL:
    case ADATA_MESG_ID_DATA_LONG:
    case SDATA_MESG_ID:
      break;

    default:
      return(-2);
    }

  if(size == 0 && deadline == 0)
    return(-3);

  /* Build complete id (ID + TYPE) */
  MesgID |= mesg_type;

  /* Check if ID/Type exists on the table */
  for(i=0;i<STA_MAX_VAR;i++)
    if(NRDB_mesgtable.mesg[i].m_id == MesgID)
      {	/* Message found; update properties */

	/* Update message size */
	if(size) 
	  { /* Update size. CAB/Com port must be removed and created again */
	    
	    if(NRDB_mesgtable.mesg[i].m_type == SDATA_MESG_ID)
	      { /* Remove the CAB associated with the synchronous message */
		cab_delete(cabset[i]);

		/* Create CAB with new size */
		sprintf(cabid,"%x",NRDB_mesgtable.mesg[i].m_id);
		cabset[i]=cab_create(cabid,NRDB_mesgtable.mesg[i].m_size*sizeof(char),3);
	      }
	    else
	      { /* Remove the Comm Port associated with the asynchronous message */
		port_delete(portset[i]);
		
		if(NRDB_mesgtable.mesg[i].m_status == PRODUCER)
		  cptype=READ; /* Station is a producer -> network layers reads the port */
		else
		  cptype=WRITE; /* Station is a consumer -> network layer writes to the port */
	      
		sprintf(portid,"%x",NRDB_mesgtable.mesg[i].m_id);
		portset[i]=port_create(portid, NRDB_mesgtable.mesg[i].m_size*sizeof(char)+sizeof(struct timespec)+\
				       sizeof(short int), STA_AM_BUFF_SIZE, STREAM, cptype);
	      } 
	  }
	
	/* Update message deadline */
	if(deadline)
	  NRDB_mesgtable.mesg[i].m_deadline=deadline;
	
	/* Return sucess */
	return(0);
      }
  
  /* If message not found, return with error */
  return(-1);
}





int NRDB_printmesg(void)
/****************************************************************************************/
/*!
\brief Prints info about the messages locally configured

   Used in stations
   Prints info about the messages locally configured (both consumed and received)

\param NRDB_tabtype: global variables' table
\param NRDB_mesgtable: global variables' table

\returns error code\n
					  0: OK\n 
*                     -1 : table empty
*****************************************************************************************/
{
  int i;
  
  /* Print the message set properties */
  cprintf("\n\n >> Message set properties:");
  for(i=0;i<STA_MAX_VAR;i++)
    if(NRDB_mesgtable.mesg[i].m_id!=0 )
    {  
      switch(NRDB_mesgtable.mesg[i].m_type)
	{
	case SDATA_MESG_ID:
	  cprintf("\n SDM            :");
	  break;
	case ADATA_MESG_ID_DATA_SHORT:
	  cprintf("\n ADM(DATA_SHORT):");
	  break;
	case ADATA_MESG_ID_ALARM:
	  cprintf("\n ADM(ALARM)     :");
	  break;
	case ADATA_MESG_ID_SIGNAL:
	  cprintf("\n ADM(SIGNAL)    :");
	  break;
	case ADATA_MESG_ID_DATA_LONG:
	  cprintf("\n ADM(DATA_LONG) :");
	  break;
	default:
	  cprintf("\n Unknown type!!!:");
	  }
      cprintf("Id=%x Len=%4u ",NRDB_mesgtable.mesg[i].m_id,NRDB_mesgtable.mesg[i].m_size);
      if( NRDB_mesgtable.mesg[i].m_status & CONSUMER )
	cprintf(" CONSUMER");
      if( NRDB_mesgtable.mesg[i].m_status & PRODUCER )
	cprintf(" PRODUCER");
    }
  
  return 0;
}



signed char NRDB_getstatus(unsigned int MesgID, WORD mesg_type,NRDB_mesgtype *mesg_prop)
/****************************************************************************************/
/*!
\brief Returns variable's status byte of message m_id

   Used in stations
   Returns variable's status byte of message m_id

\param MesgID: message id
\param mesg_type: message type
\param *mesg_prop: message properties
\param NRDB_tabtype: global variables' table
\param NRDB_mesgtable: global variables' table

\returns status byte
\returns structure with complete message info (size,deadline, #of tx/rx,  etc )
\returns error code\n
                      -1 : message not found in table\n
                      -2 : invalid type

******************************************************************************************/
{
  int i;
  signed char ret_val;
  
  /* Check for message type validity */
  switch(mesg_type)
    {
    case ADATA_MESG_ID_DATA_SHORT:
    case ADATA_MESG_ID_ALARM:
    case ADATA_MESG_ID_SIGNAL:
    case ADATA_MESG_ID_DATA_LONG:
    case SDATA_MESG_ID:
      break;

    default:
      return(-2);
    }
  
  /* Scan for message */
  MesgID |= mesg_type; 
  for(i=0;i<STA_MAX_VAR;i++)
    if(NRDB_mesgtable.mesg[i].m_id == MesgID)
      {
	/* MesgID exists in the local table, return status byte */
	ret_val = ((NRDB_mesgtable.mesg[i]).m_status & 0x7f);
	
	/* Store message info in functions's argument */
	*mesg_prop=NRDB_mesgtable.mesg[i];

	/* Exit function */
	return(ret_val);
      }
  
  /* If prog. execution reaches this point, mesg was not found in table */
  return(-1);
}






/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/  



    /*****************************************************/
   /*****************************************************/
  /*    Communication primitives                       */
 /*****************************************************/
/*****************************************************/

signed char produce (unsigned int MesgID, unsigned char *mesg_value)
/***************************************************************************/
/*! 
\brief Copies a message's value into the value buffer in the Station's Message Table

    Used in stations\n
    Copies a message's value into the value buffer in the Station's Message Table\n
    Checks wether station is a producer of this var\n
    Returns the promptness status

\param MesgID: message id
\param *mesg_value: pointer to the value to produce
\param NRDB_tabtype: global variables' table
\param NRDB_mesgtable: global variables' table

\returns promptness status of the variable (previous TX_NEW_DAT flag status )

\returns error code\n
*                     -1: mesg not found in table\n
*                     -2: station is not producer
*
***************************************************************************/
{
  int i;                  /* Counter used to scan the variables table */
  signed char ret_val;	  /* Function exit code */
  NRDB_mesgtype *aux_var; /* Pointer to particular message being produced */
  char *pcabdata;         /* Pointer to CAB buffer */

  /* Set message type header */
  MesgID |= SDATA_MESG_ID;
  
  /* Check if message exists in table */
  for(i=0;i<STA_MAX_VAR;i++)
    if(NRDB_mesgtable.mesg[i].m_id==MesgID)
      break;

  /* If message was not found, return with error */
  if(i==STA_MAX_VAR)
    return(-1);

  /* Checks is station is a producer, and if so update data */
  aux_var = &(NRDB_mesgtable.mesg[i]);

  if( (aux_var->m_status) & PRODUCER )
    {
      /* Station is producer of this var */
      pcabdata=cab_reserve(cabset[i]);	/* Get pointer for CAB buffer and reserve it */
      memcpy(pcabdata,mesg_value,aux_var->m_size*sizeof(char));	/* Copy data to CAB buffer */
      cab_putmes(cabset[i],pcabdata);		/* Release CAB buffer */

      /* Save previous TX_NEW_DAT bit */
      ret_val = (aux_var->m_status) & TX_NEW_DAT;

      /* Set the TX_NEW_DAT bit */
      (aux_var->m_status) |= TX_NEW_DAT;

      /* Update counters & status variables*/
      (aux_var->m_nprod)++; /* One more instance produced */
      aux_var->m_t2deadline=floor((aux_var->m_deadline * EC_LEN) / DDL_RES_US); /* Store initial deadline value (in SLOTS) */
      sys_gettime(&(aux_var->m_tupdate)); /* Store update instant */
      
      // cprintf("\n Prod: m_t2deadline=%d time=%lu",aux_var->m_t2deadline, TIMESPEC2USEC(&(aux_var->m_tupdate)));

    }
  else
    /* Station is not a producer of this var */
    ret_val = -2;


  return (ret_val);
}



signed char consume (unsigned int MesgID, unsigned char *mesg_value, long *deadline)
/****************************************************************************/
/*!
\brief Copies a message's value from the value buffer in the Station's Message Table

    Used in stations\n
    Copies a message's value from the value buffer in the Station's Message Table\n
    Checks wether station is a consumer of this var\n
    Returns the promptness and refreshness status

\param MesgID: message id
\param *mesg_value: pointer where to copy the value to
\param NRDB_tabtype: global variables' table
\param NRDB_mesgtable: global variables' table

\returns deadline : time to (if > 0) or after (if < 0) message time validity\n 
           This value is in usecs, but with resolution DDL_RES_US
\returns promptness and refreshness status of the variable
\returns error code\n
                    -1: var not in table\n
                    -2: station is not consumer

*****************************************************************************/
{

  int i;		 /* Counter used to scan the variables table */
  signed char ret_val=0; /* Function exit code */
  NRDB_mesgtype *aux_var;/* Pointer to particular message being produced */
  char *pcabdata;	 /* Pointer to CAB buffer */
  
  /* To handle message deadline */
  struct timespec curtime, diftime;
  TIME diftime_usec;

 /* Set message type header */
  MesgID |= SDATA_MESG_ID;

  /* Check if message exists in table */
  for(i=0;i<STA_MAX_VAR;i++)
    if(NRDB_mesgtable.mesg[i].m_id==MesgID )
      break;

  /* If message was not found, return with error */
  if(i==STA_MAX_VAR)
    return(-1);

  /* Checks is station is a consumer; if so return data */
  aux_var = &(NRDB_mesgtable.mesg[i]);

  if( (aux_var->m_status) & CONSUMER )
    {
      /* Save previous RX_NEW_DAT and TX_NEW_DAT bits */
      ret_val = (aux_var->m_status) & (RX_NEW_DAT|TX_NEW_DAT);
      /* Reset the RX_NEW_DAT bit */
      (aux_var->m_status) &= ~RX_NEW_DAT;
      
      /* Get last data from CAB structure */
      pcabdata=cab_getmes(cabset[i]);		/* Get pointer for CAB buffer with last data and reserve it*/
      memcpy(mesg_value,pcabdata,aux_var->m_size*sizeof(char));	/* Copy data to CAB buffer */
      cab_unget(cabset[i],pcabdata);		/* Release CAB buffer */

      /* Update counters & status variables*/
      (aux_var->m_ncons)++; /* One more consume action */
      sys_gettime(&curtime); 
      SUBTIMESPEC(&curtime,&aux_var->m_tupdate,&diftime);
      diftime_usec=TIMESPEC2USEC(&diftime);
      *deadline = (long)(aux_var->m_t2deadline) * DDL_RES_US - diftime_usec; /* Compute time to deadline */
      
      
    }
  else   /* Station is not a consumer of this var */
    ret_val = -2;
  
  return (ret_val);
}



int ftt_send(unsigned int MesgID, WORD mesg_type, unsigned char *mesg_value, WORD mesg_dlc)
/*******************************************************************************************/
/*!
\brief Send ftt message

   Used in stations. This function is not directly used by application software.\n
     For normal asynchronous messages : send(id,type,*value)\n
     For long asynchronous messages   : send_l(id,type,*value,dlc)\n

   Checks wether station is a producer of this var\n
   Places the data in a fifo (txmition is triggered by the comm network)

\param MesgID: message id
\param mesg_type: message type
\param mesg_value: pointer to the value to produce
\param mesg_dlc: message DLC      
\param NRDB_tabtype: global variables' table
\param NRDB_mesgtable: global variables' table

\returns error code\n
					 0 : OK\n
				    -1 : mesg not found in table\n
				    -2 : station is not sender\n
					-3 : invalid message type\n
					-4 : FIFO is FULL\n
					-5 : SIGNAL message required to send CBS tx request not found\n 
					-99 : unspecified error

***************************************************************************/
{
  int i;                     /* Counters used to scan the variables table */
  signed char ret_val;	     /* Function exit code */
  NRDB_mesgtype *aux_var;    /* Pointer to particular message being produced */
  PORT pwrite_id;            /* Port ID */
  unsigned char portid[16];  /* String with the ID of the com. port */

  /* Vars used to handle deadline info */
  struct timespec *tupdate=NULL;
  short int       *t2deadline=NULL;

  /* DATA_LONG variables */
  unsigned char dl_flag=0;
  WORD mesg_size, nmesg_dlc;

  /* Check for message type validity */
  switch(mesg_type)
    {
    case ADATA_MESG_ID_DATA_SHORT:
    case ADATA_MESG_ID_ALARM:
    case ADATA_MESG_ID_SIGNAL:
      break;
    case ADATA_MESG_ID_DATA_LONG:
      dl_flag=1; /* Signal that it is an async long data message */
      break;

    default:
      return(-3);
    }

  /* Set message type header */
  MesgID |= mesg_type;
  
  /* Check if message exists in table */
  for(i=0;i<STA_MAX_VAR;i++)
    if(NRDB_mesgtable.mesg[i].m_id==MesgID)
      break;

  /* If message was not found, return with error */
  if(i==STA_MAX_VAR)
    return(-1);

  /* Checks is station is a sender, and if so store data in the com port*/
  aux_var = &(NRDB_mesgtable.mesg[i]);

  if( (aux_var->m_status) & PRODUCER )
    {
      /* Station is a sender of this var                             */
      /* Note: the port is created by the addmesg function           */
      /* For PRODUCERS the port is created in READ mode              */
      /* For CONSUMERS the port is created in WRITE mode             */
      /* This way, the port is ready for access by the network layer */ 

      /* Connect to the port */
      sprintf(portid,"%x",aux_var->m_id);

      //cprintf("\n [ftt_send] portid=%s",portid);

      /* Compute the port data length */
      mesg_size = aux_var->m_size + sizeof(struct timespec) + sizeof(short int);
      if(dl_flag)
	mesg_size+=sizeof(WORD);
	
      pwrite_id=port_connect(portid, mesg_size, STREAM, WRITE);
      if(pwrite_id < 0)
	{
	  cprintf("\[ftt_send] Fatal error! port_connect fail");
	  return -99;
	}
      
      /* Calculate and store deadline info in send buffer*/
      tupdate=(struct timespec *)SEND_mesgbuf;
      sys_gettime(tupdate);
      t2deadline=(short int *)&SEND_mesgbuf[sizeof(struct timespec)];
      *t2deadline=floor((aux_var->m_deadline * EC_LEN) / DDL_RES_US); /* Store initial deadline value (in SLOTS) */

      /* Store dlc (if long asynch message) and message data in send buffer */
      if(dl_flag)
	{

	  //cprintf("\n[ftt_send]: long asynch mesg size=%u",mesg_dlc);
	  nmesg_dlc=htons(mesg_dlc);
	  memcpy(&SEND_mesgbuf[sizeof(struct timespec)+sizeof(short int)],\
		 (char *)(&nmesg_dlc), sizeof(WORD));	      

	  memcpy(&SEND_mesgbuf[sizeof(struct timespec)+sizeof(short int)+sizeof(WORD)],\
	  	 mesg_value,  mesg_dlc);	      

	}
      else
	memcpy(&SEND_mesgbuf[sizeof(struct timespec)+sizeof(short int)],\
	       mesg_value,  aux_var->m_size);	      

      /* Send to FIFO (deadline info + mesg data) */
      ret_val=port_send(pwrite_id,SEND_mesgbuf,NON_BLOCK);

      /* Disconnect and check exit code */
      port_disconnect(pwrite_id);

      if(!ret_val)
	return(-4);

      (aux_var->m_nprod)++; /* One more instance sent (application level) */

      /* Signal success */
      ret_val = 0;
      
      /* Print debug info */
      // cprintf("\n Send: time=%lu m_t2deadline=%d ", TIMESPEC2USEC(tupdate), *t2deadline);

    }
  else
    /* Station is not a producer of this var */
    ret_val = -2;


  return (ret_val);
}


int send(unsigned int MesgID, WORD mesg_type, unsigned char *mesg_value)
/*******************************************************************************************/
/*!
\brief Interface for ftt_send()

   Interface for ftt_send(). Used to send regular asynchronous messages
********************************************************************************************/
{
  return (ftt_send(MesgID,mesg_type,mesg_value,0));
}


int send_l (unsigned int MesgID, WORD mesg_type, unsigned char *mesg_value, WORD mesg_dlc)
/************************************************************************************/
/*!
\brief Sends long asynchronous messages using CBS servers

   Used in stations\n
   Sends long asynchronous messages using CBS servers\n
   To send this kind of messages it is required to:\n
          - Put the message in the fifo, ready to send\n
          - Send a signal message with the request to the CBS server residing on the Master\n

   !NOTE!: this function is just an interface to the standard send() function

***************************************************************************/
{
  /* Local variables */
  int status,i;
  
  /* Vars required by CBS servers */
  WORD CBS_req_data[3];
  WORD ReqMesgID;


  //cprintf("\n send_l: input args (MesgID=%x mesg_type=%x dlc=%u",MesgID,mesg_type,mesg_dlc);
  
  /* Check data type */
  if(mesg_type != ADATA_MESG_ID_DATA_LONG)
    return -3;

  //cprintf("\n   > Now sending long data message ...");
  
  /* First send the data message */
  status=ftt_send(MesgID, mesg_type, mesg_value, mesg_dlc);
  if(status)
    {
      cprintf(" [send_l] Error! Status=%x",status);
      return status; /* On error return */
    }
  
  //cprintf(" OK!");

  /* Prepare data field of SIGNAL message */
  CBS_req_data[0] = htons(SIGNAL_CBS_REQ_TYPE); /* Signal type is CBS request */;
  CBS_req_data[1] = htons(MesgID);
  CBS_req_data[2] = htons(mesg_dlc);

  /* Find SIGNAL message associated with the CBS server */
  MesgID |= ADATA_MESG_ID_DATA_LONG;
  for(i=0;i<STA_MAX_VAR;i++)
    if(NRDB_mesgtable.mesg[i].m_id==MesgID)
      break;

  if(i==STA_MAX_VAR) 
    {
      cprintf("\n [send_l]*** Strange error: long asynchronous data message not found in message table!");
    }

  ReqMesgID = NRDB_mesgtable.mesg[i].m_cbs_smid & MESGID_MASK;;
    
  //cprintf("\n   > Now sending signal message (ReqMesgID=%x)...",ReqMesgID);
  
  /* Send CBS tx request SIGNAL message */
  status=ftt_send(ReqMesgID, ADATA_MESG_ID_SIGNAL, (char *)CBS_req_data, 0);

  if(status)
    cprintf(" [send_l] Error! Status=%x",status);
    
  /* Exit function */
  return status;

}  



int ftt_receive (unsigned int MesgID, WORD mesg_type,unsigned char *mesg_value,\
	     WORD *mesg_dlc, long *deadline)
/***************************************************************************/
/*!
\brief receive ftt message

   Used in stations. This function is not used dierctly by the application software.\n
   There are 2 functions:\n
        receive()   : reception of regular asynch mesgs\n
        receive_l() : reception of long asynch messages (includes an extra field (DLC))\n
  
   Checks wether station is a receiver of this var\n
   Gets data from a receiving fifo and extracts deadline info\n

\param MesgID: message id
\param mesg_type: message type
\param mesg_dlc: message DLC
\param NRDB_tabtype: global variables' table
\param NRDB_mesgtable: global variables' table

\returns *mesg_value : message data)
\returns *deadline   : time to (if >0) or after (if < 0) deadline 
                         in usec, but with resolution DDL_RES_US
\returns error code\n
						>0 : OK \n
						0 : FIFO is Empty\n
						-1 : mesg not found in table\n
						-2 : station is not sender\n
					    -3 : invalid message type

***************************************************************************/
{
  int i;                  /* Counter used to scan the variables table */
  signed char ret_val;	  /* Function exit code */
  NRDB_mesgtype *aux_var; /* Pointer to particular message being produced */
  PORT pread_id;         /* Port ID */
  
  /* Useb by long asynch. mesgs */
  unsigned char dl_flag = 0;
  WORD mesg_size;
  
  
  unsigned char portid[16];  /* String with the ID of the com. port */

  struct timespec *tupdate=NULL;
  short int *t2deadline=NULL;

  struct timespec curtime,diftime;
  TIME diftime_usec;


  /* Check for message type validity */
  switch(mesg_type)
    {
    case ADATA_MESG_ID_DATA_SHORT:
    case ADATA_MESG_ID_ALARM:
    case ADATA_MESG_ID_SIGNAL:
      break;
    case ADATA_MESG_ID_DATA_LONG:
      dl_flag = 1;
      break;

    default:
      return(-3);
    }

  /* Set message type header */
  MesgID |= mesg_type;
  
  /* Check if message exists in table */
  for(i=0;i<STA_MAX_VAR;i++)
    if(NRDB_mesgtable.mesg[i].m_id==MesgID)
      break;

  /* If message was not found, return with error */
  if(i==STA_MAX_VAR)
    return(-1);

  /* Checks is station is a consumer, and if so get the data */
  aux_var = &(NRDB_mesgtable.mesg[i]);

  if( (aux_var->m_status) & CONSUMER )
    {
      /* Station is a consumer of this var                           */
      /* Note: the port is created by the addmesg function           */
      /* For PRODUCERS the port is created in READ mode              */
      /* For CONSUMERS the port is created in WRITE mode             */
      /* This way, the port is ready for access by the network layer */ 

      /* Connect to the port */
      sprintf(portid,"%x",aux_var->m_id);

      if(dl_flag)
	mesg_size = aux_var->m_size*sizeof(char) + sizeof(WORD);
      else
	mesg_size = aux_var->m_size*sizeof(char);

      pread_id=port_connect(portid,mesg_size+sizeof(struct timespec)+sizeof(short int),\
				     STREAM, READ);

     /* Read FIFO (deadline info + mesg data) */
      ret_val=port_receive(pread_id,RECEIVE_mesgbuf,NON_BLOCK);
      
      /* Disconnect and check exit code */
      port_disconnect(pread_id);
      if(!ret_val)
	return(0);

      /* Get and calculate deadline info from receive buffer*/
      tupdate=(struct timespec *)RECEIVE_mesgbuf;
      t2deadline=(short int *)&RECEIVE_mesgbuf[sizeof(struct timespec)];

      /* Update counters & status variables*/
      (aux_var->m_ncons)++; /* One more reception */
      sys_gettime(&curtime); 
      SUBTIMESPEC(&curtime,tupdate,&diftime);
      diftime_usec=TIMESPEC2USEC(&diftime);
      *deadline = (long)(*t2deadline) * DDL_RES_US - diftime_usec; /* Compute time to deadline */

      /* Get message data from RECEIVE buffer */
      if(dl_flag)
	{
	  memcpy((char *)mesg_dlc,&RECEIVE_mesgbuf[sizeof(struct timespec)+sizeof(short int)],\
	       sizeof(WORD));
	  *mesg_dlc=ntohs(*mesg_dlc);
	  memcpy(mesg_value,&RECEIVE_mesgbuf[sizeof(struct timespec)+sizeof(short int)+sizeof(WORD)],\
		 *mesg_dlc*sizeof(char));
	}
      else
	memcpy(mesg_value,&RECEIVE_mesgbuf[sizeof(struct timespec)+sizeof(short int)],\
	       aux_var->m_size*sizeof(char));

      /* Signal sucess */
      ret_val = 1;

      /* Print debug info */
      //cprintf("\n Receive: time=%lu deadline=%lu",TIMESPEC2USEC(&curtime),*deadline);
    }
  else
    /* Station is not a consumer of this var */
    ret_val = -2;

  return (ret_val);
}


int receive (unsigned int MesgID, WORD mesg_type,unsigned char *mesg_value,long *deadline)
/*******************************************************************************************/
/*!
\brief Consumes regular asynchronous messages

   Consumes regular asynchronous messages.\n 
   In this case the dlc field of ftt_receive() is not used since the
     messages have fixed sizes, defined during set-up.
*******************************************************************************************/
{
  WORD lixo;

  return(ftt_receive (MesgID, mesg_type,mesg_value, &lixo, deadline));

}

int receive_l (unsigned int MesgID, WORD mesg_type, unsigned char *mesg_value, \
	       WORD *mesg_dlc, long *deadline)
/***************************************************************************/
/*!
\brief Consumes long asynchronous messages.

	Consumes long asynchronous messages.	n
   In this case all the ftt_receive() fields are used
   This interface is only implemented to make the system more coherent	n
       send    / send_l	n
       receive / receive_l
      
*******************************************************************************************/
{
  
  return(ftt_receive (MesgID, mesg_type,mesg_value, mesg_dlc, deadline));

}






/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/  



    /*****************************************************/
   /*****************************************************/
  /*    FTT-protocol  functions                        */
 /*****************************************************/
/*****************************************************/


signed char ftt_get2tx_sdata(unsigned int MesgID, unsigned char *mesg_value, WORD *sdm_len, BYTE *mesg_seqn, short int *mesg_deadline, PID *pid_callback)
/*********************************************************************************************************/
/*!
\brief Gets the most recent variable's value from the respective CAB

   Used in stations\n
   Gets the most recent variable's value from the respective CAB\n
   Checks wether station is a producer of this var\n
   Returns and updates the TX_NEW_DAT flag

\param MesgID: message id
\param NRDB_tabtype: global variables' table
\param NRDB_mesgtable: global variables' table

\returns *mesg_value: buffer where data is stored\n 
\returns *sdm_len: data size\n
\returns *mesg_seqn: message sequence number\n
\returns *mesg_deadline: message time to deadline (in SLOTs)\n
\returns *pid_callback: pid of the callback task (activated on network TX)\n
\returns return value:\n
           - if >= 0 value of the TX_NEW_DAT flag\n
           - if <  0 error code\n
                       -1 : mesg not in table\n
                       -2 : station is not consumer

*********************************************************************************************************/
{
  int i;		  /* Counter used to scan the variables table */
  signed char ret_val;	  /* Function exit code */
  NRDB_mesgtype *aux_var; /* Pointer to particular message being produced */
  char *pcabdata;	  /* Pointer to CAB buffer */

  /* To handle message deadline */
  struct timespec curtime, diftime;
  TIME diftime_usec;

  /* Check if message exists in table */
  for(i=0;i<STA_MAX_VAR;i++)
    if(NRDB_mesgtable.mesg[i].m_id==MesgID )
      break;

  /* If message was not found, return with error */
  if(i==STA_MAX_VAR)
    return(-1);

  /* Checks is station is a producer */
  aux_var = &(NRDB_mesgtable.mesg[i]);

  if( (aux_var->m_status) & PRODUCER )
    {
      /* Save TX_NEW_DAT bit */
      ret_val = (aux_var->m_status) & TX_NEW_DAT;
      /* Reset the TX_NEW_DAT bit */
      (aux_var->m_status) &= ~TX_NEW_DAT;

      /* Get last data from CAB structure */
      pcabdata=cab_getmes(cabset[i]);		/* Get pointer for CAB buffer with last data and reserve it*/
      memcpy(mesg_value,pcabdata,aux_var->m_size*sizeof(char));	/* Copy data to CAB buffer */
      cab_unget(cabset[i],pcabdata);		/* Release CAB buffer */
            
      /* Update counters & status variables*/
      (aux_var->m_ntx)++; /* Message will be txmitted */
      sys_gettime(&curtime); 
      SUBTIMESPEC(&curtime,&aux_var->m_tupdate,&diftime);
      diftime_usec=TIMESPEC2USEC(&diftime);
      

      /* Fill function arguments */
      *sdm_len = aux_var->m_size;
      *mesg_seqn=aux_var->m_ntx;
      *mesg_deadline=aux_var->m_t2deadline - ceil(diftime_usec / DDL_RES_US); /* Compute time to deadline */;
      *pid_callback = aux_var->m_pid_callback;

      //cprintf("TX:Ct=%lu ddl=%d[%x] ",TIMESPEC2USEC(&curtime),*mesg_deadline,(short int)(*mesg_deadline));
      
    }
  else   /* Station is not a consumer of this var */
    ret_val = -2;
  
  return (ret_val);
}


signed char ftt_rx2update_sdata(unsigned int MesgID, unsigned char *mesg_value, WORD mesg_flags, short int mesg_t2deadline)
/****************************************************************************/
/*!
\brief Updates the CAB contents relative to a specific message with data passed as argument

   Used in stations\n
   Updates the CAB contents relative to a specific message with data passed as argument\n
   Checks wether station is a consumer of this var\n
   Returns and updates the RX_NEW_DAT flag

\param MesgID: message id
\param *mesg_value: pointer to data
\param mesg_flags: message flags
\param mesg_t2deadline: value of time to deadline field carried in the message 
\param NRDB_tabtype: global variables' table
\param NRDB_mesgtable: global variables' table

\returns
      - value of the old RX_NEW_DAT flag\n
      - error code\n
                    -1 : mesg not in table\n
                    -2 : station is not producer

*****************************************************************************/
{

  int i;	              /* Counter used to scan the variables table */
  signed char ret_val;	      /* Function exit code */
  NRDB_mesgtype *aux_var;     /* Pointer to particular message being produced */
  char *pcabdata;	      /* Pointer to CAB buffer */
  unsigned char txnewdat_bit; /* Value of the TXNEWDAT bit of the received message */
  int mesg_bits;

  /* Store txnewdat flag */
  txnewdat_bit=MesgID & TXNEWDAT_MASK;

  /* Hide TXNEWDAT flag */
  MesgID &= ~TXNEWDAT_MASK;

  /* Check if message exists in table */
  for(i=0;i<STA_MAX_VAR;i++)
    if(NRDB_mesgtable.mesg[i].m_id==MesgID )
      break;

  /* If message was not found, return with error */
  if(i==STA_MAX_VAR)
    return(-1);

  /* Checks is station is a consumer */
  aux_var = &(NRDB_mesgtable.mesg[i]);

  if( (aux_var->m_status) & CONSUMER )
    {
      /* Save RX_NEW_DAT bit */
      ret_val = (aux_var->m_status) & RX_NEW_DAT;
      
      /* Set the RX_NEW_DAT bit */
      (aux_var->m_status) |= RX_NEW_DAT;
      
      /* Set TX_NEW_DAT status flag*/
      if(txnewdat_bit)
	(aux_var->m_status) |= TX_NEW_DAT;
      else
	(aux_var->m_status) &= ~TX_NEW_DAT;

      /* Store fresh data in CAB */
      pcabdata=cab_reserve(cabset[i]);	/* Get pointer for CAB buffer and reserve it */
      memcpy(pcabdata,mesg_value,aux_var->m_size*sizeof(char));	/* Copy data to CAB buffer */
      cab_putmes(cabset[i],pcabdata);	/* Release CAB buffer */

      /* Update counters & status variables*/
      (aux_var->m_nrx)++; /* Message has been received */
      /* Calculate message size in bits */
      if(aux_var->m_size > (ETH_DATALEN_MIN - sizeof(FTT_SDM_HEADER)))
	mesg_bits=aux_var->m_size*8;
      else
	mesg_bits=ETH_DATALEN_MIN*8;
      mesg_bits += ETH_OVERHEAD_bits + FTT_IFS_bits + sizeof(FTT_SDM_HEADER)*8;

      /* Update time validity and rx time */
      aux_var->m_t2deadline = mesg_t2deadline - ceil((T2DEADLINE_OVERHEADS_US+mesg_bits*BIT_TIME_US) / DDL_RES_US);
      sys_gettime(&aux_var->m_tupdate); 
      aux_var->m_flags = mesg_flags;

      /* Activate callback task (if defined) */
      if(aux_var->m_pid_callback > 0)
	task_activate(aux_var->m_pid_callback);
              
    }
  else   /* Station is not a consumer of this var */
    ret_val = -2;
  
  return (ret_val);
}



signed char ftt_get2tx_adata(unsigned int MesgID, unsigned char *mesg_value, WORD *adm_len, BYTE *mesg_seqn, short int *mesg_deadline, PID *pid_callback)
/*********************************************************************************************************/
/*!
\brief  Gets the most recent variable's value from the respective FIFO

   Used in stations\n
   Gets the most recent variable's value from the respective FIFO\n
   Checks wether station is a sender of this var

\param MesgID: message id
\param NRDB_tabtype: global variables' table
\param NRDB_mesgtable: global variables' table

\returns *mesg_value: buffer where data is stored 
\returns *adm_len: data size
\returns *mesg_seqn: message sequence number
\returns *mesg_deadline: message time to deadline (in SLOTs)
\returns *pid_callback: pid of task to be activated (after message TX)
\returns return value\n
           - if >  0 : OK\n
           - if == 0 : FIFO empty\n
           - if <  0 : error code\n
						          -1 : mesg not in table\n
                                  -2 : station is not sender

*********************************************************************************************************/
{
  int i;		  /* Counter used to scan the variables table */
  signed char ret_val;	  /* Function exit code */
  NRDB_mesgtype *aux_var; /* Pointer to particular message being produced */
  
  /* To handle message deadline */
  struct timespec curtime, diftime;
  TIME diftime_usec;

  struct timespec *tupdate;
  short int *t2deadline;

  /* Flag to signal if a long asynch message */
  unsigned char dl_flag=0;
  WORD mdlc=0, *pmdlc;


  /* Check if message exists in table */
  for(i=0;i<STA_MAX_VAR;i++)
    if(NRDB_mesgtable.mesg[i].m_id==MesgID )
      break;

  /* If message was not found, return with error */
  if(i==STA_MAX_VAR)
    return(-1);

  /* Check if a long asynch message */
  if( (MesgID & MESGTYPE_MASK) == ADATA_MESG_ID_DATA_LONG)
    dl_flag = 1;

  /* Checks is station is a sender */
  aux_var = &(NRDB_mesgtable.mesg[i]);

  if( (aux_var->m_status) & PRODUCER )
    {
      
      /* Get data from FIFO */
      ret_val=port_receive(portset[i],FTT_SEND_mesgbuf,NON_BLOCK);
      if(!ret_val)
	return(0);
      ret_val = 1;

      /* Get pointer to deadline info */
      tupdate=(struct timespec *)FTT_SEND_mesgbuf;
      t2deadline=(short int *)&FTT_SEND_mesgbuf[sizeof(struct timespec)];
          
      /* Copy data to output buffer */
      if(dl_flag)
	{
	  pmdlc=(WORD *)&FTT_SEND_mesgbuf[sizeof(struct timespec)+sizeof(short int)];
	  mdlc=*pmdlc;
	  mdlc=ntohs(mdlc);
	  
	  //cprintf("\n[get2tx_adata] dlc=%d",mdlc);


	  /* If long asynch message, WORD with the dlc is put after the message data and the */
	  /*    numer of data bytes sent is the specified in the dlc */

	  memcpy(mesg_value, &FTT_SEND_mesgbuf[sizeof(struct timespec)+sizeof(short int)],\
		 mdlc+sizeof(WORD));    

	}
	else
	  memcpy(mesg_value, &FTT_SEND_mesgbuf[sizeof(struct timespec)+sizeof(short int)],\
		 aux_var->m_size*sizeof(char));    

      /* Update counters & status variables*/
      (aux_var->m_ntx)++; /* Message will be txmitted */
      sys_gettime(&curtime); 
      SUBTIMESPEC(&curtime,tupdate,&diftime);
      diftime_usec=TIMESPEC2USEC(&diftime);
      
      /* Fill function arguments */
      if(dl_flag)
	*adm_len = mdlc+sizeof(WORD);
      else	  
	*adm_len = aux_var->m_size;
      *mesg_seqn=aux_var->m_ntx;
      *mesg_deadline=*t2deadline - ceil(diftime_usec / DDL_RES_US); /* Compute time to deadline */;
      *pid_callback = aux_var->m_pid_callback;

      //cprintf("\n[get2tx_adata]: mesg effective size=%u",*adm_len);

      //cprintf("TX:Ct=%lu ddl=%d[%x] ",TIMESPEC2USEC(&curtime),*mesg_deadline,(short int)(*mesg_deadline));
      
    }
  else   /* Station is not a sender of this var */
    ret_val = -2;
  
  return (ret_val);
}


signed char ftt_rx2update_adata(unsigned int MesgID, unsigned char *mesg_value, WORD mesg_flags, short int mesg_t2deadline)
/****************************************************************************/
/*!
\brief Updates the FIFO contents relative to a specific asynchronous

   Used in stations\n
   Updates the FIFO contents relative to a specific asynchronous message 
        with data passed as argument\n
   Checks wether station is a consumer of this var

\param MesgID: message id
\param *mesg_value: pointer to data
\param mesg_flags: message flags
\param mesg_t2deadline: value of time to deadline field carried in the message 
\param NRDB_tabtype: global variables' table
\param NRDB_mesgtable: global variables' table

\returns pid of the callback task
\returns return value: 
          - if > 0  : FIFO Full\n
           -if == 0 : OK\n
           -if <  0\n          
                     -1 : mesg not in table\n
                     -2 : station is not receiver
                     
*****************************************************************************/
{

  int i;	              /* Counter used to scan the variables table */
  signed char ret_val;	      /* Function exit code */
  NRDB_mesgtype *aux_var;     /* Pointer to particular message being produced */
  int mesg_bits;

  /* Variables to handle deadline info */
  struct timespec *tupdate=NULL;
  short int *t2deadline=NULL;

  /* To handle long asynch messages */
  unsigned char dl_flag=0;

 
  /* Check if message exists in table */
  for(i=0;i<STA_MAX_VAR;i++)
    if(NRDB_mesgtable.mesg[i].m_id==MesgID )
      break;

  /* If message was not found, return with error */
  if(i==STA_MAX_VAR)
    return(-1);

  /* Check if a long asynch message */
  if( (MesgID & MESGTYPE_MASK) == ADATA_MESG_ID_DATA_LONG)
    dl_flag = 1;
  
  /* Checks if station is a receiver */
  aux_var = &(NRDB_mesgtable.mesg[i]);

  if( (aux_var->m_status) & CONSUMER )
    {
      

      /* Calculate message size in bits (to compute tx time in the Ethernet bus) */
      if(aux_var->m_size > (ETH_DATALEN_MIN - sizeof(FTT_ADM_HEADER)) )
	mesg_bits=(aux_var->m_size)*8;
      else
	mesg_bits=ETH_DATALEN_MIN*8;
      mesg_bits += ETH_OVERHEAD_bits + FTT_IFS_bits + sizeof(FTT_ADM_HEADER)*8;
      if(dl_flag)
	mesg_bits+= sizeof(WORD)*8;
      
      /* Update time validity and rx time */
      tupdate=(struct timespec *)FTT_RECEIVE_mesgbuf;
      t2deadline=(short int *)&FTT_RECEIVE_mesgbuf[sizeof(struct timespec)];
      sys_gettime(tupdate);
     
      *t2deadline = mesg_t2deadline - ceil((T2DEADLINE_OVERHEADS_US+mesg_bits*BIT_TIME_US) / DDL_RES_US);

      /* Store message data into receive buffer */
      if(dl_flag)
	memcpy(&FTT_RECEIVE_mesgbuf[sizeof(struct timespec)+sizeof(short int)],\
	       mesg_value,aux_var->m_size*sizeof(char)+sizeof(WORD));    
      else
	memcpy(&FTT_RECEIVE_mesgbuf[sizeof(struct timespec)+sizeof(short int)],\
	       mesg_value,aux_var->m_size*sizeof(char));    
      
      /* Store buffer in fifo (deadline info + message data) */
      ret_val = port_send(portset[i],FTT_RECEIVE_mesgbuf,NON_BLOCK);

      /* Check for success */
      if(!ret_val)
	return(1);
      ret_val = 0;
      
      /* Update counters & status variables*/
      (aux_var->m_nrx)++; /* Message has been received */

      //cprintf(" ** rx2update_adata -> activate callback **");

      /* Activate callback task (if defined) */
      if(aux_var->m_pid_callback > 0)
	task_activate(aux_var->m_pid_callback);

      /* Print debug info */
      //cprintf("\n RX: time=%lu deadline=%d",TIMESPEC2USEC(tupdate), mesg_t2deadline);
      
              
    }
  else   /* Station is not a receiver of this var */
    ret_val = -2;
  
  return (ret_val);
}




signed char ftt_test_producer(unsigned int MesgID)
/****************************************************************************/
/*!
\brief Tests if a station is a producer of the message

   Used in stations\n
   Tests if a station is a producer of the message with the ID passed in the arguments

\param MesgID: message id (Includes type field)
\param NRDB_tabtype: global variables' table
\param NRDB_mesgtable: global variables' table
  
\returns 
      - 0 : station is a producer of the message\n
      - error code\n
                    -1: var not in table\n
                    -2: station is not producer

*****************************************************************************/
{

  int i;                 /* Counter used to scan the variables table */
  signed char ret_val=0; /* Function exit code */

  /* Check if message exists in table */
  for(i=0;i<STA_MAX_VAR;i++)
    if(NRDB_mesgtable.mesg[i].m_id==MesgID )
      break;

  /* If message was not found, return with error */
  if(i==STA_MAX_VAR)
    return(-1);

  if( NRDB_mesgtable.mesg[i].m_status & PRODUCER )
    ret_val = 0;
  else   /* Station is not a producer of this var */
    ret_val = -2;
  
  return (ret_val);
}


signed char ftt_test_consumer(unsigned int MesgID)
/****************************************************************************/
/*!
\brief Tests if a station is a consumer of the message

   Used in stations\n
   Tests if a station is a consumer of the message with the ID passed in the arguments

\param MesgID: message id
\param NRDB_tabtype: global variables' table
\param NRDB_mesgtable: global variables' table
  
\returns 
      - 0 : station is a consumer of the message\n
      - error code\n
                    -1: var not in table\n
                    -2: station is not consumer
*****************************************************************************/
{

  int i;                /* Counter used to scan the variables table */
  signed char ret_val;	/* Function exit code */

  /* Hide TXNEWDAT_MASK bit */
  MesgID &= ~TXNEWDAT_MASK;

  /* Check if message exists in table */
  for(i=0;i<STA_MAX_VAR;i++)
    if(NRDB_mesgtable.mesg[i].m_id==MesgID )
      break;

  /* If message was not found, return with error */
  if(i==STA_MAX_VAR)
    return(-1);

  if( NRDB_mesgtable.mesg[i].m_status & CONSUMER )
    ret_val = 0;
  else   /* Station is not a producer of this var */
    ret_val = -2;
  
  return (ret_val);
}


  /************************************************************************/
 /*                     FTT-Ethernet Tasks                               */
/************************************************************************/

void ftt_ethproc(void *eth_pkt)
/****************************************************************************/
/*!
\brief This function waits for the reception of an Ethernet Frame

	This function waits for the reception of an Ethernet Frame.\n
	Depending on the frame type several actions are performed:\n

	- FTT Trigger Message: get IDs and lengths of the distinct messages
		to be produced and activate the production task;\n
	- Data message : update the local image value of the value of the var;\n

	Note: this function must be binded to a given protocol (frame type)!

*****************************************************************************/

{

  /****************/
  /* Declarations */
  /****************/
  FTT_TM_PKT *ftt_tm_pkt;
  WORD frame_type,frame_id,nsm,nam;
  
  BYTE *ftt_pkt_byte; /* Pointers to Ethernet frame received */ 
  WORD *ftt_pkt_word;

  int i,tm_dbind; /* General counters */

  short int frame_t2deadline; /* Time to deadline carried in the message received (for data messages) */
  WORD frame_flags;      /* Message flags data field */

  /********/
  /* Code */
  /********/

  /* Get pointer to frame just received */
  ftt_pkt_word=(WORD *)eth_getFDB(eth_pkt);

  /* Get ftt frame type */
  frame_type=(ntohs(*ftt_pkt_word) & MESGTYPE_MASK);
  
  switch(frame_type)
    {
      
    case MST_MESG_ID:  
      /*********************************/
      /***** Trigger Message received? */
      /*********************************/      

      /* Get a time stamp for the reception of the trigger message */
      sys_gettime(&tm_timestamp);
      
      /* Get pointer to TM data contents */
      ftt_tm_pkt=(FTT_TM_PKT *)eth_getFDB(eth_pkt);
      
      /* Get number of sync. and asynch. messages in the EC */
      nsm=ntohs(ftt_tm_pkt->TM_Header.TM_nsm);
      nam=ntohs(ftt_tm_pkt->TM_Header.TM_nam);
      
      /* Store messages to produce in table */
      /* 1: First synchronous messages (and asynchronous via aperiodic server) */
      tm_dbind=0;
      mesg2p_table.nsm=0;
      for(i=0;i<nsm;i++)
        {
	  mesg2p_table.mesg2p[i].msg_type = SDATA_MESG_ID;
	  mesg2p_table.mesg2p[i].msg_id   = (ftt_tm_pkt->TM_Data[tm_dbind] << 8) | (ftt_tm_pkt->TM_Data[tm_dbind+1]);
	  mesg2p_table.mesg2p[i].msg_len  = ftt_tm_pkt->TM_Data[tm_dbind+2];	  
	  tm_dbind+=3;
  	  mesg2p_table.nsm++;
        }

      /* 2: then asynchronous messages */
      mesg2p_table.nam=0;
      for(i=nsm;i<(nsm+nam);i++)
        {
	  mesg2p_table.mesg2p[i].msg_type = !(SDATA_MESG_ID);
	  mesg2p_table.mesg2p[i].msg_id   = (ftt_tm_pkt->TM_Data[tm_dbind] << 8) | (ftt_tm_pkt->TM_Data[tm_dbind+1]);
	  mesg2p_table.mesg2p[i].msg_len  = ceil(FTT_MINLEN_bits * BIT_TIME_US / SLOTV_US); /* Note: Asynch mesgs have at most 46 data bytes */
	  tm_dbind+=2;
  	  mesg2p_table.nam++;
        }
      
      /* Activate the task that will send the messages (at appropriate time instants) to the ETh bus */
      task_activate(pid_ftt_sdispatcher);

      /* Update debug var */
      dbg_eccnt++;
  
      /* End of processing */
      break;
    
      
    case SDATA_MESG_ID: 
      /******************************************/
      /***** Synchronous Data Message received  */
      /******************************************/

      /* Get ftt frame header info */
      /* Synch Data Header ID(WORD) + Flags(WORD) + t2deadline(WORD) */ 
      frame_id         = ntohs(*ftt_pkt_word); 
      frame_flags      = ntohs(*(ftt_pkt_word+1));
      frame_t2deadline = ntohs(*(ftt_pkt_word+2));
            
      if(ftt_test_consumer(frame_id)<0)
	{
	  /* Station is not a consumer of the message - do nothing */
	  break;
	}
      else
	{
	  /* Get pointer to frame data just received */
	  ftt_pkt_byte=(BYTE *)eth_getFDB(eth_pkt);
	  ftt_pkt_byte += sizeof(FTT_SDM_HEADER); /* Data follows SDM's Header */
	  ftt_rx2update_sdata(frame_id,ftt_pkt_byte,frame_flags,frame_t2deadline);
	}
	
      /* End of processing */
      break;


    
      /******************************************/
      /**** Asynchronous Data Message received  */
      /******************************************/
    case ADATA_MESG_ID_DATA_SHORT: 
    case ADATA_MESG_ID_ALARM:
    case ADATA_MESG_ID_SIGNAL:
    case ADATA_MESG_ID_DATA_LONG:
      /* Get ftt frame header info */
      /* Asynch Data Header ID(WORD) + Flags(WORD) + t2deadline(WORD) */ 
      frame_id         = ntohs(*ftt_pkt_word); 
      frame_flags      = ntohs(*(ftt_pkt_word+1));
      frame_t2deadline = ntohs(*(ftt_pkt_word+2));
            
      if(ftt_test_consumer(frame_id)<0)
	{
	  /* Station is not a consumer of the message - do nothing */
	  break;
	}
      else
	{
	  /* Get pointer to frame data just received */
	  ftt_pkt_byte=(BYTE *)eth_getFDB(eth_pkt);
	  ftt_pkt_byte += sizeof(FTT_ADM_HEADER); /* Data follows ADM's Header */
	  ftt_rx2update_adata(frame_id,ftt_pkt_byte,frame_flags,frame_t2deadline);

	  //cprintf("** ADMreceived ID=%x **",frame_id);
	}
	
      /* End of processing */
      break;
      
    } /* End Switch(frametype) */
  
}


/*************************************************************************************/

void ftt_txwakeup(void *par)
{
  /*
  struct timespec cc;
  
  cc = ll_gettime(TIME_EXACT,NULL);
  cprintf(" Event ftt_txwakeup: %lu",TIMESPEC2USEC(cc));
  */
  internal_sem_post(&sem);
}

TASK ftt_sdispatcher(void *arg)
/*******************************************************************************/
/*!
\brief This task is responsible by timely produce the station messages.

	This task is responsible by timely produce the station messages.
    When an Ethernet frame with a Trigger message arrives, the routine\n
	that handles the Ethernet frame processing fills a data structure\n
	with the IDs and release time of the messages to be produced by this station.\n
	When the Trig mesg is completely processed, it activates this task.

*****************************************************************************/
{

  /****************/
  /* Declarations */
  /****************/

  ETH_ADDR broadcast,nulladdr; /* Ethernet addresses */
  ETH_ADDR myEthAddr;

  void *TxBuff;	       /* Pointer to a Eth Transmission buffer */
  
  WORD *mesg_header;   /* Pointer to the header of a sync data message */
  BYTE *mesg_data;     /* Pointer to data of a synchronous data message */
  WORD mesg_efflen;    /* Real data length of message (in bytes) */ 
  BYTE mesg_seqn;      /* Message sequence number */
  short int  mesg_deadline;  /* Message t2deadline field */
    
  WORD MesgID; 	       /* Message ID */
  BYTE MesgLen;	       /* Message Length (in slots) */

  PID pid_callback;    /* PID of task to activate upon message transmission */

  int m2p_tabind;     /* Messages_to_produce table index */ 

  unsigned long twait;	  /* Time that must elapse prior to mesg txmition */
  signed char mesgstatus; /* Status ofa given message */

  struct timespec nexttxtime,nextwakeuptime,curtime;

  unsigned int txtrig_offset;  /* Task is waked up earlyer by an amount of txtrig_offset */
                               /* and then busy waits tuntil the exact time              */
                               /* (to compnesate some odd delays that appear sometimes)  */

  /********/
  /* Code */
  /********/
  
  /* Semaphor init */
  txtrig_offset= TXTRIG_OFFSET;
  internal_sem_init(&sem,0);

  /* Ethernet inits */  
  eth_str2Addr("FF:FF:FF:FF:FF:FF",&broadcast);
  eth_str2Addr("00:00:00:00:00:00",&nulladdr);
  eth_getAddress(&myEthAddr);
  if( (TxBuff = netbuff_get(&Pool_TxBuff,NON_BLOCK)) == NULL)
    {
      perror("[ftt_sdispatcher] Fatal Error : failed to allocate netbuff_get() returned NULL");
      sys_abort(1);
    }
  eth_setHeader(TxBuff,broadcast,ETH_FTT_TYPE);
    
  /* Variable's inits */
  twait = SCAN_WINDOW;
  m2p_tabind=0;

  while(1)
    {

      /* Check if all messages processed */
      if(m2p_tabind >= (mesg2p_table.nsm+mesg2p_table.nam))
	{
	 task_endcycle();	/* Current EC-Schedule processed; wait for next TM */
	 twait=SCAN_WINDOW;
	 m2p_tabind=0;
	 //cprintf("\n TM=%lu",TIMESPEC2USEC(&tm_timestamp));
	}

      /* Get message ID and Length */
      MesgID = mesg2p_table.mesg2p[m2p_tabind].msg_id;
      MesgLen = mesg2p_table.mesg2p[m2p_tabind].msg_len;
      //cprintf("%x",MesgID);
      
      /* Checks for errors and txmits data */
      mesgstatus=ftt_test_producer(MesgID);
      if(mesgstatus < 0)
	{
	  /* Station does not produce this message, */
	  /* but time required by it must be accounted for */
	  twait += (unsigned long)MesgLen*SLOTV_US;
	  m2p_tabind++;
	}
      else 
	{
	  /*** If Station is a producer/sender of the message          ***/
	  /*     1 - Prepare the message                                 */
	  /*     2 - Test is Tx instant far away (if so, sleep in a sem) */
	  /*     3 - Busy wait for the time of tx start                  */
	  /*     4 - Transmitt the message                               */
	  /*     5 - Update the accumulated time variable                */
	  /*   Sleeps until the end of the transmission Eth frames not produced by this station */
	  /*   Delays are relative to the reception instant of the trigger message              */

	  /***** Justification:*/ 
	  /* When using the task_delay there is a lot of jitter (100's of usecs) */
	  /* Thats because when the task is awakened there is some OS stuff to be done */
	  /* (reschedule and so on). To avoid this, an event is set for some instant before the */
	  /* right time, and then the task busy wait for the correct instant, with the preemption */
	  /* disabled.  */

	  /* 1 - Prepare the message */
	  /* Set up message frame */
	  /* Data frames have an header with 6 bytes ( ID(2)+Flags(2)+Deadline(2) ) */
	  mesg_header= (WORD *)eth_getFDB(TxBuff);
	  mesg_data  = (BYTE *)eth_getFDB(TxBuff)+3*sizeof(WORD);
 
	  if((MesgID & MESGTYPE_MASK) == SDATA_MESG_ID)
	    { /* If synchronous message */
	      mesgstatus=ftt_get2tx_sdata(MesgID,mesg_data,&mesg_efflen,&mesg_seqn,\
					  &mesg_deadline, &pid_callback); /* Get data into net buffer */
	      if(mesgstatus < 0)
		{	      
		  cprintf("\n *** [ftt_sdispatcher] Strange Error! ftt_get2tx_sdata failed after ftt_test_producer");
		  exit(1);	      
		}

	      if(mesgstatus)   /* Update the TXNEWDAT bit on the message identifier */
		MesgID |= TXNEWDAT_MASK;
	    }
	  else
	    {

	      //cprintf("+");
	      mesgstatus=ftt_get2tx_adata(MesgID,mesg_data,&mesg_efflen,&mesg_seqn,\
					  &mesg_deadline, &pid_callback); /* Get data into net buffer */
	      if(mesgstatus < 0)
		{	      
		  cprintf("\n *** [ftt_sdispatcher] Strange Error! ftt_get2tx_adata failed after ftt_test_producer");
		  exit(1);	      
		}

	      //cprintf("\n [ftt_sdispatcher] Async mesg_eff_len = %d", mesg_efflen);
	      	 
     	      /* Check if any message waiting for TX */
	      if(mesgstatus == 0)
		{		
		  /* Add the message TX time to the accum time for next mesg */
		  twait+=MesgLen*SLOTV_US;
	      
		  /* Update table index - process next message */
		  m2p_tabind++;

		  continue;
		}
	    }

	  *(mesg_header)  = htons((WORD)MesgID); /* Setup message ID */
	  *(mesg_header+1)= htons(mesg_seqn);    /* Setup Flags field */
	  *(mesg_header+2)= htons(mesg_deadline);/* Setup deadline field */

	  /* 2 - Waits for time-to-txmit */
	  /* Calculate the instant of activation of task (tx_time - offset) */
	  /* If the time is far away, set an event and wait,  else busy wait! */
	  TIMESPEC_ASSIGN(&nexttxtime,&tm_timestamp);
	  TIMESPEC_ASSIGN(&nextwakeuptime,&tm_timestamp);
	  ADDUSEC2TIMESPEC(twait,&nexttxtime);
	  ADDUSEC2TIMESPEC((twait-txtrig_offset),&nextwakeuptime);
	  //cprintf("[I=%x w=%lu t=%lu",MesgID,TIMESPEC2USEC(&nextwakeuptime),TIMESPEC2USEC(&nexttxtime));
	  sys_gettime(&curtime);	
	  //cprintf(" c=%lu]",TIMESPEC2USEC(&curtime));  

	  if(TIMESPEC_A_GT_B(&curtime,&nexttxtime))
	    txact_late++;

	  if(TIMESPEC_A_GT_B(&curtime,&nextwakeuptime))
	    txact_jit++;

	  if(TIMESPEC_A_LT_B(&curtime,&nextwakeuptime))
	    txact_early++;

	  //if(TIMESPEC_A_LT_B(&curtime,&nextwakeuptime)) /* If current time < busy wait window set event */
	  //  {
	  //     proc_table[exec_shadow].delay_timer = kern_event_post(&nextwakeuptime, ftt_txwakeup, NULL);
	  //    internal_sem_wait(&sem);
	  //  }
	  	  
	  /* 3 - Busy waits the arrival of correct instant */ 
	  for (;;) 
	    {
	      sys_gettime(&curtime);
	      if (!(TIMESPEC_A_LT_B(&curtime,&nextwakeuptime))) break;
	    } 
	  kern_cli();
	  for (;;) 
	    {
	      sys_gettime(&curtime);
	      if (!(TIMESPEC_A_LT_B(&curtime,&nexttxtime))) break;
	    } 
      
	  /* 4 - Send packet. Note that only the data field len of the Eth frame is supplied */
	  eth_sendPkt(TxBuff,3*sizeof(WORD)+mesg_efflen*sizeof(BYTE));
	  kern_sti();
	  
	  /* 5 - Activate callback (if defined) */
	  if(pid_callback)
	    task_activate(pid_callback);
	  	  
	  /* 6 - Add the message TX time to the accum time for next mesg */
	  twait+=MesgLen*SLOTV_US;

	  /* 7 - Update table index - process next message */
	  m2p_tabind++;

	  /* 8 - Debug var update */
	  dbg_txcnt++;
 	}
    }
}



/*<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/  


    /*****************************************************/
   /*****************************************************/
  /*           ftt boot functions                      */
 /*****************************************************/
/*****************************************************/

int ftt_ninit()
/*************************************************************************/
/*!
\brief This function is responsible for initializing the ftt system

    This function is responsible for initializing the ftt system:\n      
            - System tasks creation (scheduler, dispatcher, ...)\n       
            - Data structures initialization\n                           
                                                                       
   and MUST BE CALLED BEFORE ANYTHING ELSE                             
*************************************************************************/
{

  /* Task models required by FTT */
  HARD_TASK_MODEL ftt_sdispatcher_tm; /* It will be registered at level 0 */
  
  /* Other variables */
  int i;
  TIME sm_prod_mit,sm_prod_wcet;
  
  //set_exchandler_grx(); /* After execption prints PID ... */

  /*********************************************/
  /* <<<<< Module organization >>>>>           */
  /* Note : there are the following modules:   */
  /*         Level 0 : EDF                     */
  /*         Level 1 : CBS (Guest on level 0)  */
  /*         Level 2 : EDF                     */
  /*         Level 3 : CBS (Guest on level 1)  */
  /*         Level 4 : RR                      */
  /*   -The task for mesg txmittion is registered at level 0 (Highest priority level) */
  /*   -The eth_init tasks at level 1 (2nd highest priority)             */
  /*   -ALL OTHER HRT TASKS at level 2                                   */
  /*   -ALL OTHER SRT TASKS at level 3                                   */
  /*   This way the dispatcher will allways preempt any other tasks when it becomes ready */
  /*      providing a more regular Trig message. The taks related with ethernet get the   */
  /*      following level of priority.         */
  /*********************************************/

  /*******************************/
  /* Create ftt_sdispatcher task */
  /*******************************/

  cprintf("\n *** EC_LEN=%lu",EC_LEN);

  sm_prod_mit=(TIME)EC_LEN-1000;
  sm_prod_wcet=(TIME)sm_prod_mit-100;
  hard_task_default_model(ftt_sdispatcher_tm);
  hard_task_def_mit(ftt_sdispatcher_tm,sm_prod_mit);
  hard_task_def_wcet(ftt_sdispatcher_tm,sm_prod_wcet);
  hard_task_def_aperiodic(ftt_sdispatcher_tm);
  hard_task_def_level(ftt_sdispatcher_tm,0);   /* Task at level 0 */ 
  
  if ((pid_ftt_sdispatcher=task_create("ftt_sdispatcher", ftt_sdispatcher, &ftt_sdispatcher_tm, NULL)) == -1)
    {
      perror("[FTTETS1.c(main)] Could not create <ftt_sdispatcher>:");
      sys_end();
    }
  cprintf("\n[main] <ftt_sdispatcher> task created. PID=%u",(unsigned int)pid_ftt_sdispatcher);
  

  /********************/
  /* Initializations  */
  /********************/

  /* Messages to be produced within the EC */
  mesg2p_table.nsm=0;
  mesg2p_table.nam=0;

  /* NRDB_mesgtable (messages handled locally) init*/
  NRDB_mesgtable.m_nmesgs=0; /* No messages in table */
  for( i=0; i < STA_MAX_VAR; i++)
          NRDB_mesgtable.mesg[i].m_id = 0;

  return 0;

}

int ftt_nstart()
/*************************************************************************/
/*!
\brief This function is responsible for starting the system (nodes)

    This function is responsible for starting the system (nodes)\n       
    It can be called before or after the message definition\n            
    If called after the task definition the system starts with\n       
       the same time reference for all the messages.\n                   
                                                                       
   Note: ftt_ninit must be called before this function                 
                                                                       
*************************************************************************/
{

  /*************/
  /* Setup net */
  /*************/
  netbuff_init(&Pool_TxBuff,2,ETH_MAXLEN_BYTES); /* Alocate tx buffers */
  eth_setProtocol(ETH_FTT_TYPE,ftt_ethproc);	/* Bind frame type Rxption with handler */

  return 0;
  
}
